import numpy as np
from numpy import sin, cos, tan, pi, sqrt
import yaml
import os
# import imp

# import welleng.error
from ..utils import NEV_to_HLA

# since this is running on different OS flavors
PATH = os.path.dirname(__file__)
FILENAME = os.path.join(
    '', *[PATH, 'tool_codes', 'error_codes.yaml']
)


class iscwsaMwd:
    def __init__(
        self,
        error,
        model
    ):
        """
        Class using the ISCWSA MWD (Rev4) model to determine well bore
        uncertatinty.

        Parameters
        ----------
            error: an intitiated welleng.error.ErrorModel object

        Returns
        -------
            errors: welleng.error.ErrorModel object
                A populated ErrorModel object for the selected error model.
        """
        error.__init__

        self.e = error
        self.errors = {}

        with open(FILENAME, 'r') as file:
            iscwsa_error_models = yaml.safe_load(file)
        self.em = iscwsa_error_models[model]
        if 'Default Tortusity (rad/m)' in self.em['header']:
            self.tortuosity = self.em['header']['Default Tortusity (rad/m)']
        else:
            self.tortuosity = None

        if model == "iscwsa_mwd_rev5":
            assert self.tortuosity is not None, (
                "No default tortuosity defined in model header"
            )

        self._initiate_func_dict()

        for err in self.em['codes']:
            # func = self._get_the_func_out(err)
            func = self.em['codes'][err]['function']
            mag = self.em['codes'][err]['magnitude']
            propagation = self.em['codes'][err]['propagation']
            self.errors[err] = (
                self.call_func(
                    code=err,
                    func=func,
                    error=self.e,
                    mag=mag,
                    propagation=propagation,
                    tortuosity=self.tortuosity,
                )
            )

        self.cov_NEVs = np.zeros((3, 3, len(self.e.survey_rad)))
        for _, value in self.errors.items():
            self.cov_NEVs += value.cov_NEV

        self.cov_HLAs = NEV_to_HLA(self.e.survey_rad, self.cov_NEVs)

    def _get_the_func_out(self, err):
        if err in self.exceptional_funcs:
            func = self.exceptional_funcs[err]
        else:
            func = self.em['codes'][err]['function']

        return func

    def call_func(self, code, func, error, mag, propagation, **kwargs):
        """
        Function for calling functions by mapping function labels to their
        functions.
        """
        assert func in self.func_dict, f"no function for function {func}"

        return self.func_dict[func](code, error, mag, propagation, **kwargs)

    def _initiate_func_dict(self):
        """
        This dictionary will need to be updated if/when additional error
        functions are added to the model.
        """
        self.func_dict = {
            'ABXY_TI1': ABXY_TI1,
            'ABXY_TI2': ABXY_TI2,
            'ABZ': ABZ,
            'AMIL': AMIL,
            'ASXY_TI1': ASXY_TI1,
            'ASXY_TI2': ASXY_TI2,
            'ASXY_TI3': ASXY_TI3,
            'ASZ': ASZ,
            'DBH': DBH,
            'AZ': AZ,
            'DREF': DREF,
            'DSF': DSF,
            'DST': DST,
            'MBXY_TI1': MBXY_TI1,
            'MBXY_TI2': MBXY_TI2,
            'MBZ': MBZ,
            'MSXY_TI1': MSXY_TI1,
            'MSXY_TI2': MSXY_TI2,
            'MSXY_TI3': MSXY_TI3,
            'MSZ': MSZ,
            'SAG': SAG,
            'XYM1': XYM1,
            'XYM2': XYM2,
            'XYM3': XYM3,
            'XYM4': XYM4,
            'SAGE': SAGE,
            'XCL': XCL,  # requires an exception
            'XYM3L': XYM3L,  # looks like there's a mistake in the ISCWSA model
            'XYM4L': XYM4L,
        }


# error functions #
def DREF(code, error, mag=0.35, propagation='random', NEV=True, **kwargs):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DSF(
    code, error, mag=0.00056, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    dpde = dpde * np.array(error.survey_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DST(
    code, error, mag=0.00000025, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.full((len(error.survey_rad), 3), [1., 0., 0.])
    dpde[:, 0] = error.survey.tvd
    dpde = dpde * np.array(error.survey_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABXY_TI1(
    code, error, mag=0.0040, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -cos(error.survey_rad[:, 1]) / error.survey.header.G
    dpde[:, 2] = (
        cos(error.survey_rad[:, 1])
        * tan(error.survey.header.dip)
        * sin(error.survey.azi_mag_rad)
    ) / error.survey.header.G
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ABXY_TI2(
    code, error, mag=0.004, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            (
                (
                    tan(-(error.survey_rad[:, 1]) + (pi/2))
                    - tan(error.survey.header.dip)
                    * cos(error.survey.azi_mag_rad)
                ) / error.survey.header.G
            ),
            posinf=0.0,
            neginf=0.0
        )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['double_delta_md']
            * -sin(error.drdp_sing['azi2']) * mag
        ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['double_delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        if error.error_model.split('_')[-1] == 'rev5':
            e_NEV_sing[1, 1] = (
                (
                    error.survey.md[2]
                    + error.survey.md[1]
                    - 2 * error.survey.md[0]
                ) / 2
                * mag * cos(error.survey.azi_true_rad[1])
                / error.survey.header.G
            )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.array(
            0.5 * error.drdp_sing['delta_md']
                * -sin(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        e = np.array(
            0.5 * error.drdp_sing['delta_md']
                * cos(error.drdp_sing['azi2']) * mag
            ) / error.survey.header.G
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        if error.error_model.split('_')[-1] == 'rev5':
            e_NEV_star_sing[1, 1] = (
                (error.survey.md[1] - error.survey.md[0])
                * mag
                * (
                    cos(error.survey.azi_true_rad[1])
                    / error.survey.header.G
                )
            )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def ABZ(code, error, mag=0.004, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = -sin(np.array(error.survey_rad)[:, 1]) / error.survey.header.G
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip) * sin(error.survey.azi_mag_rad)
    ) / error.survey.header.G
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI1(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(
        np.array(error.survey_rad)[:, 1]
    ) * cos(np.array(error.survey_rad)[:, 1]) / sqrt(2)
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * -tan(error.survey.header.dip) * cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / sqrt(2)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI2(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(
        np.array(error.survey_rad)[:, 1]
    ) * cos(np.array(error.survey_rad)[:, 1]) / 2
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * -tan(error.survey.header.dip) * cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASXY_TI3(
    code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip) * cos(error.survey.azi_mag_rad)
        - cos(np.array(error.survey_rad)[:, 1])) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def ASZ(code, error, mag=0.0005, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * cos(np.array(error.survey_rad)[:, 1])
    )
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * tan(error.survey.header.dip)
        * cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBXY_TI1(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / (error.survey.header.b_total * cos(error.survey.header.dip))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBXY_TI2(
    code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        cos(error.survey.azi_mag_rad)
        / (
            error.survey.header.b_total
            * cos(error.survey.header.dip)
        )
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MBZ(code, error, mag=70.0, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
    ) / (error.survey.header.b_total * cos(error.survey.header.dip))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSXY_TI1(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
        * (
            tan(error.survey.header.dip)
            * cos(np.array(error.survey_rad)[:, 1])
            + sin(np.array(error.survey_rad)[:, 1])
            * cos(error.survey.azi_mag_rad)
        ) / sqrt(2)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSXY_TI2(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        sin(error.survey.azi_mag_rad) * (
            tan(error.survey.header.dip)
            * sin(np.array(error.survey_rad)[:, 1])
            * cos(np.array(error.survey_rad)[:, 1])
            - cos(np.array(error.survey_rad)[:, 1])
            * cos(np.array(error.survey_rad)[:, 1])
            * cos(error.survey.azi_mag_rad) - cos(error.survey.azi_mag_rad)
        ) / 2
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSXY_TI3(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        cos(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad) * cos(error.survey.azi_mag_rad)
        - cos(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad) * sin(error.survey.azi_mag_rad)
        - tan(error.survey.header.dip) * sin(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad)
    ) / 2
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def MSZ(
    code, error, mag=0.0016, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = -(
        sin(np.array(error.survey_rad)[:, 1])
        * cos(error.survey.azi_mag_rad)
        + tan(error.survey.header.dip) * cos(np.array(error.survey_rad)[:, 1])
    ) * sin(np.array(error.survey_rad)[:, 1]) * sin(error.survey.azi_mag_rad)
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def AZ(code, error, mag=0.00628, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DBH(
    code, error, mag=np.radians(5000), propagation='systematic', NEV=True,
    **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1 / (
        error.survey.header.b_total * cos(error.survey.header.dip)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def DBHR(
    code, error, mag=np.radians(3000), propagation='random', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = 1 / (
        error.survey.header.b_total * cos(error.survey.header.dip)
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def AMIL(code, error, mag=220.0, propagation='systematic', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = (
        -sin(np.array(error.survey_rad)[:, 1])
        * sin(error.survey.azi_mag_rad)
        / (error.survey.header.b_total * cos(error.survey.header.dip))
    )
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def SAG(
    code, error, mag=0.00349, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(np.array(error.survey_rad)[:, 1])
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def SAGE(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = sin(np.array(error.survey_rad)[:, 1]) ** 0.25
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM1(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = np.absolute(sin(np.array(error.survey_rad)[:, 1]))
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM2(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    propagation = 'systematic'  # incorrect in the rev5 model tab
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 2] = -1
    e_DIA = dpde * mag

    return error._generate_error(code, e_DIA, propagation, NEV)


def XYM3(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = (
        np.absolute(cos(np.array(error.survey_rad)[:, 1]))
        * cos(error.survey.azi_true_rad)
    )
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            -(
                np.absolute(cos(np.array(error.survey_rad)[:, 1]))
                * sin(error.survey.azi_true_rad)
            ) / sin(np.array(error.survey_rad)[:, 1]),
            posinf=0.0,
            neginf=0.0
        )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.array(0.5 * error.drdp_sing['double_delta_md'] * mag)
        e = np.zeros(len(error.drdp_sing['double_delta_md']))
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.array(0.5 * error.drdp_sing['delta_md'] * mag)
        e = np.zeros(len(error.drdp_sing['delta_md']))
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XYM4(
    code, error, mag=0.00175, propagation='systematic', NEV=True, **kwargs
):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[:, 1] = np.absolute(
        cos(np.array(error.survey_rad)[:, 1])
    ) * sin(error.survey.azi_true_rad)
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            (
                np.absolute(np.cos(np.array(error.survey_rad)[:, 1]))
                * cos(error.survey.azi_true_rad)
            )
            / sin(np.array(error.survey_rad)[:, 1]),
            posinf=0.0,
            neginf=0.0
            )
    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        n = np.zeros(len(error.drdp_sing['double_delta_md']))
        e = np.array(0.5 * error.drdp_sing['double_delta_md'] * mag)
        v = np.zeros_like(n)
        e_NEV_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        n = np.zeros(len(error.drdp_sing['delta_md']))
        e = np.array(0.5 * error.drdp_sing['delta_md'] * mag)
        v = np.zeros_like(n)
        e_NEV_star_sing = np.vstack(
            (
                np.zeros((1, 3)),
                np.stack((n, e, v), axis=-1),
                np.zeros((1, 3))
            )
        )
        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XCL(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    """
    Dummy function to manage the ISCWSA workbook not correctly defining the
    weighting functions.
    """
    tortuosity = kwargs['tortuosity']
    if code == "XCLA":
        return XCLA(
            code, error, mag=mag, propagation=propagation, NEV=NEV,
            tortuosity=tortuosity
        )
    else:
        return XCLH(
            code, error, mag=mag, propagation=propagation, NEV=NEV,
            tortuosity=tortuosity
        )


def XCLA(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))

    def manage_sing(error, kwargs):
        temp = np.absolute(
            sin(error.survey.inc_rad[1:])
            * (((
                error.survey.azi_true_rad[1:]
                - error.survey.azi_true_rad[:-1]
                + pi
            ) % (2 * pi)) - pi)
        )
        temp[np.where(
            error.survey.inc_rad[:-1] < error.survey.header.vertical_inc_limit
        )] = 0
        return temp

    dpde[1:, 0] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            manage_sing(error, kwargs),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * -sin(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 1] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            manage_sing(error, kwargs),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.azi_true_rad[1:])
    )

    e_DIA = dpde * mag

    return error._generate_error(
        code, e_DIA, propagation, NEV, e_NEV=e_DIA, e_NEV_star=e_DIA
    )


def XCLH(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[1:, 0] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.inc_rad[1:])
        * cos(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 1] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * cos(error.survey.inc_rad[1:])
        * sin(error.survey.azi_true_rad[1:])
    )

    dpde[1:, 2] = (
        (error.survey.md[1:] - error.survey.md[0:-1])
        * np.amax(np.stack((
            np.absolute(
                (error.survey.inc_rad[1:] - error.survey.inc_rad[:-1])
            ),
            (
                kwargs['tortuosity']
                * (error.survey.md[1:] - error.survey.md[0:-1])
            )
        ), axis=-1), axis=-1)
        * -sin(error.survey.inc_rad[1:])
    )

    e_DIA = dpde * mag

    return error._generate_error(
        code, e_DIA, propagation, NEV, e_NEV=e_DIA, e_NEV_star=e_DIA
    )


def XYM3L(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    coeff = np.ones(len(error.survey.md) - 1)
    coeff = np.amax(np.stack((
        coeff,
        sqrt(
            10 / (error.survey.md[1:] - error.survey.md[:-1])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey_rad), 3))
    dpde[1:, 1] = np.absolute(
        cos(error.survey.inc_rad[1:])
        * cos(error.survey.azi_true_rad[1:])
        * coeff
    )
    dpde[0, 1] = dpde[1, 1]

    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[1:, 2] = np.nan_to_num(
            (
                -np.absolute(
                    cos(error.survey.inc_rad[1:])
                )
                * (
                    sin(error.survey.azi_true_rad[1:])
                    / sin(error.survey.inc_rad[1:])
                )
                * coeff
            ),
            posinf=0,
            neginf=0
        )

    dpde[0, 2] = dpde[1, 2]

    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[1:-1, 0] = (
            coeff[:-1]
            * (
                error.survey.md[2:]
                - error.survey.md[:-2]
            ) / 2
            * mag
        )
        e_NEV_sing[1, 0] = (
            coeff[1]
            * (
                error.survey.md[2] + error.survey.md[1]
                - 2 * error.survey.md[0]
            ) / 2
            * mag
        )
        e_NEV_sing[-1, 0] = (
            coeff[-1]
            * (
                error.survey.md[-1]
                - error.survey.md[-2]
            ) / 2
            * mag
        )

        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV)
        e_NEV_star_sing[1:, 0] = (
            (
                error.survey.md[1:]
                - error.survey.md[:-1]
            ) / 2
            * mag
        )

        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )


def XYM4L(code, error, mag=0.0167, propagation='random', NEV=True, **kwargs):
    propagation = 'random'
    coeff = np.ones(len(error.survey.md))
    coeff[1:] = np.amax(np.stack((
        coeff[1:],
        sqrt(
            10 / (error.survey.md[1:] - error.survey.md[:-1])
        )
    ), axis=-1), axis=-1)

    dpde = np.zeros((len(error.survey_rad), 3))
    with np.errstate(divide='ignore', invalid='ignore'):
        dpde[:, 2] = np.nan_to_num(
            np.absolute(
                cos(error.survey.inc_rad)
                * cos(error.survey.azi_true_rad)
                / sin(error.survey.inc_rad)
                * coeff
            ),
            posinf=0,
            neginf=0,
        )

    dpde[:, 1] = (
        np.absolute(
            cos(error.survey.inc_rad)
        )
        * (
            sin(error.survey.azi_true_rad)
        )
        * coeff
    )

    e_DIA = dpde * mag

    sing = np.where(
        error.survey_rad[:, 1] < error.survey.header.vertical_inc_limit
    )
    if len(sing[0]) < 1:
        return error._generate_error(code, e_DIA, propagation, NEV)
    else:
        e_NEV = error._e_NEV(e_DIA)
        e_NEV_sing = np.zeros_like(e_NEV)
        e_NEV_sing[1:-1, 1] = (
            coeff[1:-1]
            * (
                error.survey.md[2:]
                - error.survey.md[:-2]
            ) / 2
            * mag
        )
        e_NEV_sing[1, 1] = (
            coeff[1]
            * (
                error.survey.md[2] + error.survey.md[1]
                - 2 * error.survey.md[0]
            ) / 2
            * mag
        )
        e_NEV_sing[-1, 1] = (
            coeff[-1]
            * (
                error.survey.md[-1]
                - error.survey.md[-2]
            ) / 2
            * mag
        )

        e_NEV[sing] = e_NEV_sing[sing]

        e_NEV_star = error._e_NEV_star(e_DIA)
        e_NEV_star_sing = np.zeros_like(e_NEV)
        e_NEV_star_sing[1:, 1] = (
            (
                error.survey.md[1:]
                - error.survey.md[:-1]
            ) / 2
            * mag
        )
        e_NEV_star_sing[1, 1] = (
            (
                error.survey.md[1]
                - error.survey.md[0]
            )
            * mag
        )

        e_NEV_star[sing] = e_NEV_star_sing[sing]

        return error._generate_error(
            code, e_DIA, propagation, NEV, e_NEV, e_NEV_star
        )



import numpy as np
from scipy.optimize import minimize
from ..version import __version__
from ..survey import export_csv as x_csv


def export_csv(
    survey, filename, tolerance=0.1, dls_cont=False, decimals=3, **kwargs
):
    """
    Wrapper for survey.export_csv

    Function to export a minimalist (only the control points - i.e. the
    begining and end points of hold and/or turn sections) survey to input into third
    party trajectory planning software.

    Parameters
    ----------
    survey: welleng.survey.Survey object
    filename: str
        The path and filename for saving the text file.
    tolerance: float (default: 0.1)
        How close the the final N, E, TVD position of the minimalist survey
        should be to the original survey point (e.g. within 1 meter)
    dls_cont: bool
        Whether to explicitly check for dls continuity. May result in a
        larger number of control points but a trajectory that is a closer
        fit to the survey.
    decimals: int (default: 3)
        Number of decimal places provided in the output file listing
    """

    return x_csv(
        survey, filename, tolerance=tolerance, dls_cont=dls_cont,
        decimals=decimals, **kwargs
    )



import os
import yaml
from ..survey import (
    TurnPoint,
    Survey,
    SurveyHeader,
    get_sections,
    from_connections,
)
from ..connector import (
    Connector,
    # interpolate_well,
    # get_survey
)
from ..survey import from_connections

try:
    import utm
    UTM = True
except ImportError:
    UTM = False

import numpy as np
from datetime import datetime
from ..version import __version__ as VERSION

# TODO: need to relocate the class Target to target.py


class Target:
    def __init__(
        self,
        name,
        location=None,
        geometry={
            'type': None,
            'locked': None,
            'offset': None,
            'orientation': None,
            'radius_1': None,
            'radius_2': None,
            'dip': None,
            'azimuth': None,
            'vertices': [],
            'thickness_up': None,
            'thickness_down': None,
            'color': {
                'color': None,
                'interpreter': None,
                'application': None,
                'feature': None
            },
            'category': None,
        },
    ):
        self.name = name
        self.location = location
        self.geometry = geometry


class SurveyPoint:
    def __init__(
        self,
        md=None,
        inc=None,
        azi=None,
        cov_xx=None,
        cov_xy=None,
        cov_xz=None,
        cov_yy=None,
        cov_yz=None,
        cov_zz=None,
        x_bias=None,
        y_bias=None,
        z_bias=None,
        tool=None,
        location=None,
    ):
        self.md = md
        self.inc = inc
        self.azi = azi
        self.cov_xx = cov_xx
        self.cov_xy = cov_xy
        self.cov_xz = cov_xz
        self.cov_yy = cov_yy
        self.cov_yz = cov_yz
        self.cov_zz = cov_zz
        self.x_bias = x_bias
        self.y_bias = y_bias
        self.z_bias = z_bias
        self.tool = tool
        self.location = location


class WellPlan:
    def __init__(
        self,
        depth_unit='meters',
        surface_unit='meters',
        survey=None,
        plan_name=None,
        parent_name=None,
        location_type=None,
        plan_method='curve_only',
        dirty_flag=None,
        sidetrack_id=None,
        dls=3.0,
        extension=0,
        wbp_data=None,
        targets=[],
        line=None,
        parent_wbp_file=None,
    ):
        """
        An object for storing data extracted from or for writing to a .wbp
        file. As such, the following parameters are driven by those
        required by Landmark's .wbp format.

        Parameters
        ----------
            depth_unit: string (default: 'meters')
                The units used for expressing depth (z axis or tvd) in either
                'meters' or 'feet'.
            surface_unit: string (default: 'meters')
                The units used for expressing lateral distances (x, y, N, E)
                in either 'meters' or 'feet'.
            survey: welleng.survey.Survey object (default: None)
            plan_name: string (default: None)
                The name of the well bore plan.
            parent_name: string (default: None)
                The name of the parent well bore plan (in the event that
                the planned well is a sidetrack or lateral).
            location_type: string (default: None)
                Best to review the wbp.yaml file for options.
            plan_method: string (default: 'curve_only')
                The method used for joining the plan points in the .wbp file.
                Options can be reviewed in the wbp.yaml file but won't
                currently effect how the code runs, so just leave default.
            dirt_flag: string (default: None)
                Again, review the wbp.yaml file for options, but this is
                not currently used in this code.
            sidetrack_id: string (default: None)
                Leave default, not used.
            dls: float (default: 0)
                Suggests that this sets the design dls for planning, but
                doesn't appear to matter so leave as default.
            extension: float (default: 0)
                Not really sure what this does.
            wbp_data: list of strings (default: None)
                A list of strings with each string representing a line from
                of text loaded from a .wbp file. Used for importing .wbp
                data.
            targets: list of welleng.exchange.wbp.Target objects (default: [])
                A list of target objects, but more of a future function.
            line: int (default: None)
                Used for processing .wbp files that contain multiple well
                bores.

        Returns
        -------
            A welleng.exchange.wbp.WellPlan object representing a well bore.
        """
        # Import the wbp.yaml file as a dictionary
        wbp_dict_file = os.path.join(
            os.path.dirname(__file__),
            'wbp.yaml')

        with open(wbp_dict_file) as f:
            self.wbp_dict = yaml.load(f, Loader=yaml.FullLoader)

        self.parent_data = get_parent_survey(parent_wbp_file)

        self.action = {
            'depth': self._get_units,
        }

        self.depth_unit = depth_unit
        self.surface_unit = surface_unit
        self.targets = targets
        if line is None:
            self.lines = 0
        else:
            self.lines = line

        self.wbp_data = wbp_data
        self.survey = survey
        if self.wbp_data is not None:
            assert self.survey is None, "Either wbp_data or survey"
            self.steps = []
            self.tie_on_point_flag = False
            self._process_wbp_data()
        else:
            assert isinstance(
                survey, Survey
            ), "Not a welleng Survey"
            assert self.wbp_data is None, "Either wbp_data or survey"
            assert plan_name is not None, "Must provide plan_name"
            assert location_type is not None, "Must provide a location type"

            self.plan_name = plan_name
            self.parent_name = parent_name
            self.location_type = str(location_type)
            self.plan_method = plan_method
            self.dirty_flag = str(dirty_flag)
            self.sidetrack_id = "" if sidetrack_id is None else sidetrack_id
            self.dls = dls
            self.extension = extension

            self.steps = get_sections(survey)

            self._get_surface_location()
            self._get_local_coordinates()

    def _get_local_coordinates(self):
        sd = np.zeros(3)
        sd[:2] = self.surface_datum[:2]
        self.env_local = np.vstack([
            s.location - sd for s in self.steps
        ])

    def _get_surface_location(self):
        if self.parent_data is not None:
            for line in self.parent_data:
                if "L:" in line:
                    code, e, n, v = line.split()
                    self.surface_datum = np.array([e, n, v]).astype(np.float)
                    break
        else:
            self.surface_datum = np.array([
                self.survey.n,
                self.survey.e,
                self.survey.tvd
            ]).T[0]

    def _process_wbp_data(self):
        """
        Steps through imported .wbp data and interprets line by line to
        populate a WellPlan object.
        """
        # TODO: finish coding the rest of the target inputs
        self.flag = None
        for i, line in enumerate(self.wbp_data):
            if i < self.lines:
                continue
            # first split the line and look for section headers
            l = line.split()
            m = line.split(':')
            if l[0] == '!':
                pass
            elif l[0] == "DEPTH":
                self._get_units(l[1])
                self.flag = None
            elif l[0] == "TARGETS:":
                self.flag = 'targets'
            elif l[0] == "WELLPLANS:":
                # pass
                self.flag = 'wellplans'
                self.tie_on_point_flag = True
            elif self.flag == 'targets':
                if m[0] == "T":
                    self._initiate_target(m[1])
                elif m[0] == "L":
                    self._add_target_location(m[1])
                elif m[0] == "C":
                    self._add_target_color(line)
                elif m[0] == "G":
                    self._add_target_geometry(line)
                # need to add the rest of the target inputs
            else:
                if m[0] == "W":
                    if self.flag == 'done':
                        break
                    else:
                        self.flag = 'done'
                        self.tie_on_point_flag = True
                        self._add_wellplan_header(line)
                elif m[0] == "P":
                    self._add_turn_point(line)
                elif m[0] == "L":
                    self._add_location_data(m[1])
                    self.tie_on_point_flag = False
                elif m[0] == "X":
                    self._add_extended_survey_point(l[1:])
                elif m[0] == "S":
                    self._add_survey_data(l[1:])
                else:
                    pass
            self.lines += 1

    def _initiate_target(self, name):
        self.targets.append(Target(name))

    def _add_target_location(self, data):
        x, y, z = data.split()
        self.targets[-1].location = [float(x), float(y), float(z)]

    def _add_target_geometry(self, data):
        tg = self.targets[-1].geometry
        tg.type = self.wbp_dict['TARGETS']['type'][data[2]]
        tg.locked = data[3]
        tg.offset = [float(data[4:13], float(data[13:23]))]

    def _add_target_color(self, data):
        tgc = self.targets[-1].geometry['color']
        if len(data) > 5:
            tgc['color'] = int(data[2:5])
            tgc['interpreter'] = data[5:10]
            tgc['application'] = data[11:21]
            tgc['feature'] = data[22:64]
        else:
            tgc['color'] = int(data[2:])

    def _add_wellplan_header(self, data):
        self.location_type = self.wbp_dict['LOCATION']['type'][data[2]]
        self.plan_method = self.wbp_dict['LOCATION']['plan_method'][data[4]]
        if data[5] in self.wbp_dict['LOCATION']['dirty_flag']:
            self.dirty_flag = self.wbp_dict['LOCATION']['dirty_flag'][data[5]]
        else:
            self.dirty_flag = None
        self.sidetrack_id = string_strip(data[14:22])
        self.plan_name = string_strip(data[23:84])
        self.parent_name = string_strip(data[84:144])    
        self.dls = string_strip(data[144:151], is_float=True)
        self.extension = string_strip(data[151:160], is_float=True)
        self.dls_kickoff = string_strip(data[160:171], is_float=True)

    def _add_turn_point(self, line):
        to = TurnPoint()
        data = line.split(':')[1].split()
        to.md = float(data[0])
        to.azi = float(data[1])
        to.inc = float(data[2])
        to.build_rate = float(data[3])
        to.turn_rate = float(data[4])
        to.dls = float(data[5])
        to.toolface = float(data[6])
        if self.tie_on_point_flag:
            to.tie_on = True
        else:
            to.method = string_strip(data[7])
            try:
                to.target = string_strip(data[8])
            except:
                to.target = None
        self.steps.append(to)

    def _add_extended_survey_point(self, data):
        xsp = SurveyPoint()
        xsp.cov_xx = float(data[0])
        xsp.cov_xy = float(data[1])
        xsp.cov_xz = float(data[2])
        xsp.cov_yy = float(data[3])
        xsp.cov_yz = float(data[4])
        xsp.cov_zz = float(data[5])
        xsp.x_bias = float(data[6])
        xsp.y_bias = float(data[7])
        xsp.z_bias = float(data[8])
        xsp.tool = " ".join(data[9:])
        self.steps.append(xsp)

    def _add_survey_data(self, data):
        s = self.steps[-1]
        s.md, s.azi, s.inc = [float(x) for x in data]

    def _add_location_data(self, data):
        x, y, z = data.split()
        self.steps[-1].location = [float(x), float(y), float(z)]

    def _get_units(self, key):
        self.depth_unit = self.wbp_dict['DEPTH'][key]['depth']
        self.surface_unit = self.wbp_dict['DEPTH'][key]['surface']


def string_strip(string, is_float=False):
    s = string.strip()
    if len(s) > 0:
        if is_float:
            return float(s)
        else:
            return s
    else:
        return None


def get_parent_survey(filename):
    if filename is None:
        return None

    assert filename[-4:] == '.wbp', 'Wrong format'
    with open(filename) as f:
        wbp_data = [line.rstrip() for line in f]

    flag = True
    data = []
    for line in wbp_data:
        if "WELLPLANS" in line:
            flag = False
            continue
        if flag:
            continue
        data.append(line)

    return data


# TODO: update so that filename can also be data
def load(filename):
    """
    Loads data line by line from a .wbp file, initiates a WellPlan object
    and populates it with data.

    Parameters
    ----------
        filename: string
            The location and filename of the .wbp file to load.

    Returns
    -------
        A welleng.exchange.wbp.WellPlan object
    """
    assert filename[-4:] == '.wbp', 'Wrong format'
    with open(filename) as f:
        wbp_data = [line.rstrip() for line in f]

    total_lines = len(wbp_data)
    line = 0

    depth_unit = None
    surface_unit = None
    targets = []

    well_plans = []

    while True:
        well_plans.append(
            WellPlan(
                wbp_data=wbp_data,
                depth_unit=depth_unit,
                surface_unit=surface_unit,
                targets=[],
                line=line
            )
        )
        w = well_plans[-1]
        line = w.lines
        if line == total_lines:
            return well_plans
        else:
            # keep passing these on to each well sequentially
            depth_unit = w.depth_unit
            surface_unit = w.surface_unit
            targets = w.targets


def add_targets(doc, targets):
    for t in targets:
        doc.append(f"T:{t.name}")
        doc = add_location(doc, t.location)
        doc.append((
            f"C:"
            f"{str(t.geometry['color']['color']).rjust(3)}"
        ))
    return doc


def add_location(doc, location):
    x, y, z = location
    doc.append((
            f"L:"
            f"{x:>13.2f}"
            f"{y:>13.2f}"
            f"{z:>11.2f}"
        ))
    return doc


def add_header(doc, data):
    d = data.wbp_dict['LOCATION']
    ignored_integer_1 = "   0"
    ignored_integer_2 = " 123"
    plan_name = "" if data.plan_name is None else data.plan_name
    parent_name = "" if data.parent_name is None else data.parent_name
    doc.append((
        f"W:"
        f"{get_key(d['type'], data.location_type)}"
        f"{get_unit_key(data)}"
        f"{get_key(d['plan_method'], data.plan_method)}"
        f"{get_key(d['dirty_flag'], data.dirty_flag)}"
        f"{ignored_integer_1}"
        f"{ignored_integer_2}"
        f" {data.sidetrack_id:>7}"
        f" {plan_name:<60}"
        f" {parent_name:<60}"
        f" {data.dls:>6.2f}"
        f" {data.extension:>7.5f}"
    ))
    # the documentation says there should be a kickoff dls at the end,
    # but I've never seen it and what does the other dls refer to?
    return doc


def add_step(doc, step):
    if isinstance(step, TurnPoint):
        doc = add_turn_point(doc, step)
    else:
        doc = add_survey_point(doc, step)
    return doc


def add_turn_point(doc, step):
    if step.tie_on:
        method = ""
    else:
        method = "0" if step.method is None else step.method
    target = "" if step.target is None else step.target
    # if the toolface is > -99 then need to chop off a decimal
    # this might also be the case with other variables... add similar logic
    # if that needs changing
    toolface = step.toolface
    toolface = f'{step.toolface:>7.2f}' if toolface < -99 else f'{toolface:>7.3f}'
    doc.append((
        f"P:"
        f" {step.md:>7.2f}"
        f" {step.azi:>7.3f}"
        f" {step.inc:>7.3f}"
        f" {step.build_rate:>7.3f}"
        f" {step.turn_rate:>7.3f}"
        f" {step.dls:>7.3f}"
        f" {toolface}"
        f" {method:>4}"
        f" {target:>10}"
    ))
    doc = add_location(doc, step.location)
    return doc


def add_survey_point(doc, step):
    doc.append((
        f"X:"
        f" {step.cov_xx:>8.1f}"
        f" {step.cov_xy:>8.1f}"
        f" {step.cov_xz:>8.1f}"
        f" {step.cov_yy:>8.1f}"
        f" {step.cov_yz:>8.1f}"
        f" {step.cov_zz:>8.1f}"
        f" {step.x_bias:>7.2f}"
        f" {step.y_bias:>7.2f}"
        f" {step.z_bias:>7.2f}"
        f" {step.tool:<38}"
    ))
    doc.append((
        f"S:"
        f" {step.md:>7.2f}"
        f" {step.azi:>7.3f}"
        f" {step.inc:>7.3f}"
    ))
    doc = add_location(doc, step.location)
    return doc


def get_unit_key(data):
    key = [
        key for key in data.wbp_dict['DEPTH'] if (
            data.wbp_dict['DEPTH'][key]['depth'] == data.depth_unit
            and data.wbp_dict['DEPTH'][key]['surface'] == data.surface_unit
        )
    ]
    return key[0]


def get_key(d, value):
    key = [
        key for key in d if (
            d[key] == value
        )
    ]
    return key[0]


def export(data, filename=None, comments=None):
    """
    Export a WellPlan object to .wbp format.

    Parameters
    ----------
        data: welleng.exchange.wbp.WellPlan object or a list of objects
        filename: string (default: None)
            The filename to save the .wbp file to. If None then the
            output is returned as data.
        comments: list of strings (default: None)
            A list of comments to be printed in the header of the .wbp
            file.

    Returns
    -------
        doc: list of strings

    """
    doc = []
    if not isinstance(data, list):
        data = [data]
    for i, w in enumerate(data):
        assert isinstance(
            w, WellPlan
        ), "Not a WellPlan object"
        if i == 0:
            doc.append(f"DEPTH {get_unit_key(w)}")
            doc = add_comments(doc, comments)
            doc.append("TARGETS:")
            doc = add_targets(doc, w.targets)
            doc.append("WELLPLANS:")
            if w.parent_data is not None:
                for line in w.parent_data:
                    doc.append(line)
        doc = add_header(doc, w)
        for s in w.steps:
            doc = add_step(doc, s)

    if filename is None:
        return doc
    else:
        save_to_file(doc, filename)


def add_comments(doc, comments):
    doc.append(f"! {datetime.now():%Y-%m-%d %H:%M:%S%z}")
    if comments is None:
        doc.append(f"! welleng v{VERSION}")
        doc.append("! Written by Jonny Corcutt")
    else:
        for c in comments:
            doc.append(f"! {c}")
    return doc


def save_to_file(doc, filename):
    with open(f"{filename}", 'w') as f:
        f.writelines(f'{l}\n' for l in doc)


def wbp_to_survey(
    data, step=None, radius=10, azi_reference='true', convergence=0.0,
    utm_zone=31, utm_north=True
):
    """
    Converts a WellPlan object created from a .wbp file into a Survey object.

    Parameters
    ----------
        data: wellend.exchange.wbp.WellPlan object
        step: float
            The desired step interval used to create the Survey object.
            e.g. step=30 would create a survey station every 30 meters.
        radius: float (default: 10)
            The radius of the well bore generated in the survey. The
            default is used assuming that the well will be rendered with
            welleng.visual.plot.

    Returns
    -------
        survey: welleng.survey.Survey object
    """
    assert UTM, "Missing utm library, try pip install welleng[easy]"
    connections = []
    for i, s in enumerate(data.steps):
        if i == 0:
            if isinstance(s, TurnPoint):
                dls = [s.dls]
            else:
                dls = [0.]
            e, n, v = s.location
            start_nev = [n, e, v * -1]
            continue
        e, n, v = s.location
        p = np.array([n, e, v * -1])  # - np.array(start_nev)
        if i == 1:
            md = [s.md]
            inc = [s.inc]
            azi = [s.azi]
            pos = [p]
            plan = [isinstance(s, TurnPoint)]
            continue

        # need to set dls_design relatively small to trigger adaption to
        # the dls used in the imported design, or to set it to the actual
        # dls used in the design.
        # TODO: update the connector code so that None can be passed for
        # dls_design, which will set np.inf for radius_design and force
        # the dls to be set by radius_critical.
        if isinstance(s, TurnPoint):
            dls_design = s.dls if s.dls > 0 else 1e-5
        else:
            # dls_design = data.dls if data.dls > 0 else None
            dls_design = 1e-5

        c = Connector(
            pos1=pos[-1],
            md1=md[-1],
            inc1=inc[-1],
            azi1=azi[-1],
            md2=s.md,
            inc2=s.inc,
            azi2=s.azi,
            dls_design=dls_design,
        )
        if isinstance(s, TurnPoint):
            dls.append(s.dls)
        else:
            dls.append(0)
        connections.append(c)
        plan.append(isinstance(s, TurnPoint))
        pos.append(c.pos_target)
        inc.append(np.degrees(c.inc_target))
        azi.append(np.degrees(c.azi_target))
        md.append(c.md_target)

    start_nev = np.array(pos[0])

    survey_data = from_connections(
        connections,
        step=step
    )

    lat, lon = utm.to_latlon(
        start_nev[1],
        start_nev[0],
        utm_zone,
        northern=utm_north,
    )

    sh = SurveyHeader(
        latitude=lat,
        longitude=lon,
        altitude=start_nev[-1] * -1,
        azi_reference=azi_reference,
        convergence=convergence,
    )

    survey = from_connections(
        survey_data,
        survey_header=sh,
        start_nev=start_nev,
        radius=radius,
        deg=False
    )

    # Because of the way the imported file is processed, there will likely
    # be duplicate survey stations in the survey. This function strips out
    # these duplicates and rebuilds the survey.
    survey = strip_duplicates(survey)

    return survey


def strip_duplicates(survey):
    """
    Function to strip out identical successive survey stations from a Survey
    object.

    Parameters
    ----------
        survey: welleng.survey.Survey object

    Returns
    -------
        survey_stripped: welleng.survey.Survey object
            A survey object with repeating survey stations removed.
    """
    temp = []
    for i, s in enumerate(zip(
        survey.md, survey.inc_rad, survey.azi_grid_rad, survey.radius
    )):
        if i == 0:
            temp.append(s)
            continue
        # if s == temp[-1]:
        if s[1] == temp[-1][1]:
            continue
        else:
            temp.append(s)

    sh = survey.header
    sh.azi_reference = 'grid'

    md, inc, azi, radius = np.array(temp).reshape(-1, 4).T

    survey_stripped = Survey(
        md=md,
        inc=inc,
        azi=azi,
        deg=False,
        start_nev=survey.start_nev,
        radius=radius,
        header=sh,
    )

    return survey_stripped



import numpy as np
from copy import copy, deepcopy
from scipy.spatial import distance
from scipy.optimize import minimize
try:
    from numba import njit
    NUMBA = True
except ImportError:
    NUMBA = False

from .utils import (
    get_vec, _get_angles, get_angles, get_nev, get_xyz, get_unit_vec,
    NEV_to_HLA, dls_from_radius
)
from .node import Node, get_node_params


class Connector:
    def __init__(
        self,
        node1=None,
        node2=None,
        pos1=[0., 0., 0.],
        vec1=None,
        inc1=None,
        azi1=None,
        md1=0,
        dls_design=3.0,
        dls_design2=None,
        md2=None,
        pos2=None,
        vec2=None,
        inc2=None,
        azi2=None,
        degrees=True,
        unit='meters',
        min_error=1e-5,
        delta_dls=0.1,
        min_tangent=0.,
        max_iterations=1_000,
        force_min_curve=False,
        closest_approach=False
    ):

        """
        A class to provide a fast, efficient method for determining well
        bore section geometry using the appropriate method based upon the
        provided parameters, with the intent of assisting machine learning
        fitness functions. Interpolation between the returned control points
        can be performed posthumously - attempts are made to keep processing to
        a minimum in the event that the Connector is being used for machine
        learning.

        Only specific combinations of input data are permitted, e.g. if you
        input both a start vector AND a start inc and azi you'll get an error
        (which one is correct after all?). Think about what you're trying to
        achieve and provide the data that will facilitate that outcome.

        Parameters
        ----------
        pos1: (3) list or array of floats (default: [0,0,0])
            Start position in NEV coordinates.
        vec1: (3) list or array of floats or None (default: None)
            Start position unit vector in NEV coordinates.
        inc1: float or None (default: None)
            Start position inclination.
        azi2: float or None (default: None)
            Start position azimuth.
        md1: float or None (default: None)
            Start position measured depth.
        dls_design: float (default: 3.0)
            The desired Dog Leg Severity (DLS) for the (first) curved
            section in degrees per 30 meters or 100 feet.
        dls_design2: float or None (default: None)
            The desired DLS for the second curve section in degrees per
            30 meters or 100 feet. If set to None then `dls_design` will
            be the default value.
        md2: float or None (default: None)
            The measured depth of the target position.
        pos2: (3) list or array of floats or None (default: None)
            The position of the target in NEV coordinates.
        vec2: (3) list or array of floats or None (default: None)
            The target unit vector in NEV coordinates.
        inc1: float or None (default: None)
            The inclination at the target position.
        azi2: float or None (default: None)
            The azimuth at the target position.
        degrees: boolean (default: True)
            Indicates whether the input angles (inc, azi) are in degrees
            (True) or radians (False).
        unit: string (default: 'meters')
            Indicates the distance unit, either 'meters' or 'feet'.
        min_error: float (default: 1e-5):
            Infers the error tolerance of the results and is used to set
            iteration stops when the desired error tolerance is met. Value
            must be less than 1. Use with caution as the code may
            become unstable if this value is changed.
        delta_radius: float (default: 20)
            The delta radius (first curve and second curve sections) used
            as an iteration stop when balancing radii. If the resulting
            delta radius yielded from `dls_design` and `dls_design2` is
            larger than `delta_radius`, then `delta_radius` defaults to
            the former.
        delta_dls: float (default: 0.1)
            The delta dls (first curve and second curve sections) used as an
            iteration stop when balancing radii, i.e. if the dls of the second
            section is within 0.1 deg/30m of the first curve section then the
            section is considered balanced and no further iterations are
            performed. Setting this value too low will likely result in hitting
            the recursion limit.
        min_tangent: float (default: 10)
            The minimum tangent length in the `curve_hold_curve` method
            used to mitigate instability during iterations (where the
            tangent section approaches or equals 0).
        max_iterations: int (default: 1000)
            The maximum number of iterations before giving up trying to
            fit a `curve_hold_curve`. This number is limited by Python's
            depth of recursion, but if you're hitting the default stop
            then consider changing `delta_radius` and `min_tangent` as
            your expectations may be unrealistic (this is open source
            software after all!)

        Results
        -------
        connector: welleng.connector.Connector object
        """
        if node1 is not None:
            pos1, vec1, md1 = get_node_params(
                node1
            )
        if node2 is not None:
            pos2, vec2, md2 = get_node_params(
                node2
            )

        # Set up a lookup dictionary to use with the logic to determine
        # what connector method to deploy. Use a binary string to
        # represent the inputs in the order:
        # (md2, inc2, azi2, pos2, vec2)
        # Initially I used boolean logic, but it quickly became non-
        # transparent and difficult to debug.
        self._get_initial_methods()

        # METHODS = [
        #     'hold',
        #     'curve_hold',
        #     'min_dist_to_target',
        #     'min_curve_to_target',
        #     'curve_hold_curve',
        #     'min_curve'
        # ]

        # quick check that inputs are workable and if not some steer to
        # the user.
        assert vec1 is not None or (inc1 is not None and azi1 is not None), (
            "Require either vec1 or (inc1 and azi1)"
        )
        if vec1 is not None:
            assert inc1 is None and azi1 is None, (
                "Either vec1 or (inc1 and azi1)"
            )
        if (inc1 is not None or azi1 is not None):
            assert vec1 is None, "Either vec1 or (inc1 and azi1)"

        assert (
            md2 is not None
            or pos2 is not None
            or vec2 is not None
            or inc2 is not None
            or azi2 is not None
        ), "Missing target parameters"

        if vec2 is not None:
            assert not (inc2 or azi2), "Either vec2 or (inc2 and azi2)"
        if (inc2 is not None or azi2 is not None):
            assert vec2 is None, "Either vec2 or (inc2 and azi2)"
        if md2 is not None:
            assert pos2 is None, "Either md2 or pos2"
            assert md2 >= md1, "md2 must be larger than md1"

        if dls_design is None:
            dls_design = 3.0
        else:
            assert dls_design > 0, "dls_design must be greater than zero"
        assert min_error < 1, "min_error must be less than 1.0"

        # figure out what method is required to connect the points
        target_input = convert_target_input_to_booleans(
            md2, inc2, azi2, pos2, vec2
        )

        self.force_min_curve = force_min_curve
        if self.force_min_curve:
            self.initial_method = 'min_curve_or_hold'
        else:
            self.initial_method = self.initial_methods[target_input]

        # do some more initialization stuff
        self.min_error = min_error
        self.min_tangent = min_tangent
        self.iterations = 0
        self.max_iterations = max_iterations
        self.errors = []
        self.radii = []
        self.dogleg_old, self.dogleg2_old = 0, 0
        self.dist_curve2 = 0
        self.pos1 = np.array(pos1)

        self.pos2, self.vec2, self.inc2, self.azi2, self.md2 = (
            None, None, None, None, None
        )

        # fill in the input data gaps
        if (inc1 is not None and azi1 is not None):
            if degrees:
                self.inc1 = np.radians(inc1)
                self.azi1 = np.radians(azi1)
            else:
                self.inc1 = inc1
                self.azi1 = azi1
            self.vec1 = np.array(get_vec(
                self.inc1, self.azi1, nev=True, deg=False
            )).reshape(3)
        else:
            self.vec1 = np.array(vec1).reshape(3)
            self.inc1, self.azi1 = get_angles(self.vec1, nev=True).reshape(2)

        self.md1 = md1
        self.pos_target = None if pos2 is None else np.array(pos2).reshape(3)
        self.md_target = md2

        if vec2 is not None:
            self.vec_target = np.array(vec2).reshape(3)
            self.inc_target, self.azi_target = get_angles(
                self.vec_target,
                nev=True
            ).reshape(2)
        elif (inc2 is not None and azi2 is not None):
            if degrees:
                self.inc_target = np.radians(inc2)
                self.azi_target = np.radians(azi2)
            else:
                self.inc_target = inc2
                self.azi_target = azi2
            self.vec_target = get_vec(
                self.inc_target, self.azi_target, nev=True, deg=False
            ).reshape(3)
        elif inc2 is None and azi2 is None:
            self.inc_target, self.azi_target, self.vec_target = (
                self.inc1, self.azi1, self.vec1
            )
        elif inc2 is None:
            self.inc_target = self.inc1
            if degrees:
                self.azi_target = np.radians(azi2)
            else:
                self.azi_target = azi2
            self.vec_target = get_vec(
                self.inc_target, self.azi_target, nev=True, deg=False
            ).reshape(3)
        elif azi2 is None:
            self.azi_target = self.azi1
            if degrees:
                self.inc_target = np.radians(inc2)
            else:
                self.inc_target = inc2
            self.vec_target = get_vec(
                self.inc_target, self.azi_target, nev=True, deg=False
            ).reshape(3)
        else:
            self.vec_target = vec2
            self.inc_target = inc2
            self.azi_target = azi2

        self.unit = unit
        if self.unit == 'meters':
            self.denom = 30
        else:
            self.denom = 100

        if degrees:
            self.dls_design = np.radians(dls_design)
            if dls_design2:
                self.dls_design2 = np.radians(dls_design2)
        else:
            self.dls_design = dls_design
            if dls_design2:
                self.dls_design2 = dls_design2
        if not dls_design2:
            self.dls_design2 = self.dls_design

        self.radius_design = self.denom / self.dls_design
        self.radius_design2 = self.denom / self.dls_design2

        self.delta_dls = delta_dls

        # some more initialization stuff
        self.tangent_length = None
        self.dogleg2 = None

        self.pos3, self.vec3, self.inc3, self.azi3, self.md3 = (
            None, None, None, None, None
        )
        self.radius_critical, self.radius_critical2 = np.inf, np.inf
        self.closest_approach = closest_approach

        # Things fall apart if the start and end vectors exactly equal
        # one another, so need to check for this and if this is the
        # case, modify the end vector slightly. This is a lazy way of
        # doing this, but it's fast. Probably a more precise way would
        # be to split the dogleg in two, but that's more hassle than
        # it's worth.
        if (
            self.vec_target is not None
            and np.array_equal(self.vec_target, self.vec1 * -1)
        ):
            (
                self.vec_target,
                self.inc_target,
                self.azi_target
            ) = mod_vec(self.vec_target, self.min_error)

        # properly figure out the method
        self._get_method()

        # and finally, actually do something...
        self._use_method()

        self._get_nodes()

    def _get_nodes(self):
        self.node_start = Node(
            pos=self.pos1.reshape(3),
            vec=self.vec1.reshape(3),
            md=self.md1
        )
        self.node_end = Node(
            pos=self.pos_target.reshape(3),
            vec=self.vec_target.reshape(3),
            md=self.md_target
        )

    def _min_dist_to_target(self):
        (
            self.tangent_length,
            self.dogleg
        ) = min_dist_to_target(self.radius_design, self.distances)
        self.dogleg = check_dogleg(self.dogleg)
        self.dist_curve, self.func_dogleg = get_curve_hold_data(
            self.radius_design, self.dogleg
        )
        self.vec_target = get_vec_target(
            self.pos1,
            self.vec1,
            self.pos_target,
            self.tangent_length,
            self.dist_curve,
            self.func_dogleg
        )
        self._get_angles_target()
        self._get_md_target()
        self.pos2 = (
            self.pos_target - (
                self.tangent_length * self.vec_target
            )
        )
        self.md2 = self.md1 + abs(self.dist_curve)
        self.md_target = self.md2 + self.tangent_length
        self.vec2 = self.vec_target

    def _min_curve_to_target(self):
        (
            self.tangent_length,
            self.radius_critical,
            self.dogleg
        ) = min_curve_to_target(self.distances)
        self.dogleg = check_dogleg(self.dogleg)
        self.dist_curve, self.func_dogleg = get_curve_hold_data(
            min(self.radius_design, self.radius_critical), self.dogleg
        )
        self.vec_target = get_vec_target(
            self.pos1,
            self.vec1,
            self.pos_target,
            self.tangent_length,
            self.dist_curve,
            self.func_dogleg
        )
        self._get_angles_target()
        self._get_md_target()

    def _use_method(self):
        if self.method == 'hold':
            self._hold()
        elif self.method == 'min_curve':
            self._min_curve()
        elif self.method == 'curve_hold_curve':
            self.pos2_list, self.pos3_list = [], [deepcopy(self.pos_target)]
            self.vec23 = [np.array([0., 0., 0.])]
            self.delta_radius_list = []
            # self._target_pos_and_vec_defined(deepcopy(self.pos_target))
            self._target_pos_and_vec_defined(
                self.pos1 + (self.pos_target - self.pos1) / 2
            )
        else:
            self.distances = self._get_distances(
                self.pos1, self.vec1, self.pos_target
            )
            if self.radius_design <= get_radius_critical(
                self.radius_design, self.distances, self.min_error
            ):
                self.method = 'min_dist_to_target'
                self._min_dist_to_target()
            else:
                if self.closest_approach:
                    self.method = 'min_curve_to_target'
                    self._closest_approach()
                else:
                    self.method = 'min_curve_to_target'
                    self._min_curve_to_target()

    def _get_method(self):
        assert self.initial_method not in [
            'no_input',
            'vec_and_inc_azi',
            'md_and_pos'
        ], f"{self.initial_method}"
        if self.initial_method == 'hold':
            self.method = 'hold'
        elif self.initial_method[-8:] == '_or_hold':
            if np.array_equal(self.vec_target, self.vec1):
                if self.pos_target is None:
                    self.method = 'hold'
                elif np.allclose(
                        self.vec_target,
                        (self.pos_target - self.pos1)
                        / np.linalg.norm(self.pos_target - self.pos1)
                ):
                    self.method = 'hold'
                else:
                    self.method = self.initial_method[:-8]
            else:
                self.method = self.initial_method[:-8]
        else:
            self.method = self.initial_method

    def _get_initial_methods(self):
        # TODO: probably better to load this in from a yaml file
        # [md2, inc2, azi2, pos2, vec2] forms the booleans
        self.initial_methods = {
            '00000': 'no_input',
            '00001': 'min_curve_or_hold',
            '00010': 'curve_hold_or_hold',
            '00011': 'curve_hold_curve_or_hold',
            '00100': 'min_curve_or_hold',
            '00101': 'vec_and_inc_azi',
            '00110': 'curve_hold',
            '00111': 'vec_and_inc_azi',
            '01000': 'min_curve_or_hold',
            '01001': 'vec_and_inc_azi',
            '01010': 'curve_hold_or_hold',
            '01011': 'vec_and_inc_azi',
            '01100': 'min_curve_or_hold',
            '01101': 'vec_and_inc_azi',
            '01110': 'curve_hold_curve_or_hold',
            '01111': 'vec_and_inc_azi',
            '10000': 'hold',
            '10001': 'min_curve_or_hold',
            '10010': 'md_and_pos',
            '10011': 'md_and_pos',
            '10100': 'min_curve_or_hold',
            '10101': 'vec_and_inc_azi',
            '10110': 'md_and_pos',
            '10111': 'md_and_pos',
            '11000': 'min_curve_or_hold',
            '11001': 'vec_and_inc_azi',
            '11010': 'md_and_pos',
            '11011': 'md_and_pos',
            '11100': 'min_curve_or_hold',
            '11101': 'vec_and_inc_azi',
            '11110': 'md_and_pos',
            '11111': 'md_and_pos'
        }

    def _closest_approach(self):
        vec_pos1_pos_target = self.pos_target - self.pos1
        vec_pos1_pos_target /= np.linalg.norm(vec_pos1_pos_target)

        cross_product = np.cross(vec_pos1_pos_target, self.vec1)
        cross_product /= np.linalg.norm(cross_product)

        factor = cross_product / vec_pos1_pos_target
        factor /= abs(factor)

        cc = (
            self.pos1 + cross_product * factor * self.radius_design
        )

        cc_pos_target = self.pos_target - cc
        cc_pos_target /= np.linalg.norm(cc_pos_target)

        self.pos_target_original = copy(self.pos_target)

        self.pos_target = cc + cc_pos_target * self.radius_design

        # recalculate self.distances with new self.pos_target
        self.distances = self._get_distances(
                self.pos1, self.vec1, self.pos_target
            )

        self._min_curve_to_target()

    def _min_curve(self):
        self.dogleg = get_dogleg(
            self.inc1, self.azi1, self.inc_target, self.azi_target
        )

        self.dogleg = check_dogleg(self.dogleg)
        if self.md_target is None:
            self.md2 = None
            self.dist_curve, self.func_dogleg = get_curve_hold_data(
                        self.radius_design, self.dogleg
                    )
            self.md_target = self.md1 + abs(self.dist_curve)
            self.pos_target = get_pos(
                    self.pos1,
                    self.vec1,
                    self.vec_target,
                    self.dist_curve,
                    self.func_dogleg
                ).reshape(3)
        else:
            with np.errstate(divide='ignore'):
                self.radius_critical = np.nan_to_num(abs(
                    (self.md_target - self.md1) / self.dogleg
                ), nan=np.inf)
            if (
                self.radius_critical > self.radius_design
                or (
                    np.around(self.dogleg, decimals=5)
                    == np.around(np.pi, decimals=5)
                )
            ):
                self.md2 = (
                    self.md1
                    + min(self.radius_design, self.radius_critical)
                    * self.dogleg
                )
                (
                    self.inc2, self.azi2, self.vec2
                ) = self.inc_target, self.azi_target, self.vec_target
                self.dist_curve, self.func_dogleg = get_curve_hold_data(
                        min(self.radius_design, self.radius_critical),
                        self.dogleg
                )
                self.pos2 = get_pos(
                    self.pos1,
                    self.vec1,
                    self.vec2,
                    self.dist_curve,
                    self.func_dogleg
                ).reshape(3)
                self.pos_target = self.pos2 + (
                    self.vec2 * (self.md_target - self.md2)
                )
            else:
                self.dist_curve, self.func_dogleg = get_curve_hold_data(
                        self.radius_critical, self.dogleg
                    )
                self.md2 = None
                self.pos_target = get_pos(
                    self.pos1,
                    self.vec1,
                    self.vec_target,
                    self.dist_curve,
                    self.func_dogleg
                ).reshape(3)

    def _hold(self):
        if self.pos_target is None:
            self.pos_target = (
                self.pos1 + self.vec1 * (self.md_target - self.md1)
            )
        if self.md_target is None:
            self.md_target = (
                np.linalg.norm(self.pos_target - self.pos1)
                + self.md1
            )
        self.dls, self.dls2 = 0.0, 0.0

    def _get_angles_target(self):
        self.inc_target, self.azi_target = get_angles(
            self.vec_target, nev=True
        ).reshape(2)

    def _get_md_target(self):
        self.md_target = (
            self.dist_curve
            + self.tangent_length
            + self.dist_curve2
            + self.md1
        )

    def _get_pos2(self, pos1, vec1, pos_target, radius):
        distances = self._get_distances(pos1, vec1, pos_target)

        radius_temp = get_radius_critical(
            self.radius_design,
            distances,
            self.min_error
        )
        if radius_temp > radius:
            radius_temp = radius
            assert self.radius_critical > 0

        (
            tangent_length,
            dogleg
        ) = min_dist_to_target(
            radius_temp,
            distances
        )

        dogleg = check_dogleg(dogleg)

        dist_curve, func_dogleg = get_curve_hold_data(
                    radius_temp,
                    dogleg
                )
        vec3 = get_vec_target(
            pos1,
            vec1,
            pos_target,
            tangent_length,
            dist_curve,
            func_dogleg
        )

        tangent_temp = self._get_tangent_temp(tangent_length)

        pos2 = (
            pos_target - (
                tangent_temp * vec3
            )
        )
        return pos2

    def _target_pos_and_vec_defined(self, pos3, vec_old=[0., 0., 0.]):
        """
        Function for fitting a curve, hold, curve path between a pair of
        points with vectors.

        It's written in this odd way to allow a solver function like scipy's
        optimize to solve it, but so far I've not had much success using
        these.

        If the curve sections can't be achieved with the design DLS values,
        this function will iterate until the two curve section DLSs are
        approximately balances (in DLS terms) within the prescribed delta_dls
        parameter.
        """
        args = (self, None, [0., 0., 0.])
        minimize_target_pos_and_vec_defined(
            *(
                ([
                    self.radius_design,
                    self.radius_design2
                ],)
                + args + (True,)
            )
        )
        if not all((
            self.radius_critical > self.radius_design,
            self.radius_critical2 > self.radius_design2
        )):
            while True and self.iterations <= self.max_iterations:
                self.radius_critical = (
                    (self.md_target - self.md1)
                    / (abs(self.dogleg) + abs(self.dogleg2))
                )
                assert self.radius_critical >= 0
                self.radius_critical2 = self.radius_critical
                args = (self, deepcopy(self.pos3), deepcopy(self.vec23[-1]))
                minimize_target_pos_and_vec_defined(
                    *(
                        ([
                            self.radius_design,
                            self.radius_design2
                        ],)
                        + args + (True,)
                    )
                )
                if abs(
                    self.dls - self.dls2
                ) < self.delta_dls:
                    break
            self._happy_finish()
            return

        self._happy_finish()

    def _happy_finish(self):
        # get pos1 to pos2 curve data
        self.dist_curve, self.func_dogleg = get_curve_hold_data(
            min(self.radius_critical, self.radius_design),
            self.dogleg
        )

        self.vec3 = get_vec_target(
            self.pos1,
            self.vec1,
            self.pos3,
            self.tangent_length,
            self.dist_curve,
            self.func_dogleg
        )

        self.vec2 = self.vec3

        self.pos2 = get_pos(
            self.pos1,
            self.vec1,
            self.vec3,
            self.dist_curve,
            self.func_dogleg
        ).reshape(3)

        self.md2 = self.md1 + abs(self.dist_curve)

        self.dist_curve2, self.func_dogleg2 = get_curve_hold_data(
            min(self.radius_critical2, self.radius_design2),
            self.dogleg2
        )

        self.pos3 = get_pos(
            self.pos_target,
            self.vec_target * -1,
            self.vec3 * -1,
            self.dist_curve2,
            self.func_dogleg2
        ).reshape(3)

        tangent_length = np.linalg.norm(
            self.pos3 - self.pos2
        )

        self.md3 = self.md2 + tangent_length

        self.md_target = self.md3 + abs(self.dist_curve2)

    def interpolate(self, step=30):
        return interpolate_well([self], step)


    def _get_tangent_temp(self, tangent_length):
        if np.isnan(tangent_length):
            tangent_temp = self.min_tangent
        else:
            tangent_temp = max(tangent_length, self.min_tangent)

        return tangent_temp

    def _mod_pos(self, pos):
        pos_rand = np.random.random(3)  # * self.delta_radius
        pos += pos_rand

    def _get_distances(self, pos1, vec1, pos_target):
        # When initializing a `curve_hold_curve` and pos_target is directly
        # ahead it can cause issues (it's a hold and not a curve). So this code
        # checks for that condition and if it's the case, will move the
        # target_pos a sufficient amount to prevent issues.
        vec_temp = np.array(pos_target - pos1)
        with np.errstate(divide='ignore', invalid='ignore'):
            vec_temp = vec_temp / np.linalg.norm(vec_temp)
        if np.array_equal(vec1, vec_temp):
            self._mod_pos(pos_target)
        if np.allclose(pos1, pos_target):
            return (0, 0, 0)

        else:
            dist_to_target = (
                np.linalg.norm((pos_target - pos1))
            )

            dist_perp_to_target = (
                np.dot((pos_target - pos1), vec1)
            )
            if dist_perp_to_target > dist_to_target:
                # since a tolerance is being used, occasionally things can go
                # wrong and need to be caught.
                dist_perp_to_target = dist_to_target

            dist_norm_to_target = (
                (
                    dist_to_target ** 2
                    - dist_perp_to_target ** 2
                ) ** 0.5
            )

            return (
                dist_to_target,
                dist_perp_to_target,
                dist_norm_to_target
            )


def minimize_target_pos_and_vec_defined(
    x, c, pos3=None, vec_old=[0., 0., 0.], result=False
):
    radius1, radius2 = x
    if pos3 is None:
        pos2_init = c._get_pos2(c.pos1, c.vec1, c.pos_target, radius1)
        pos3_init = c._get_pos2(
            c.pos_target, c.vec_target * -1, c.pos1, radius2
        )
        c.pos3 = pos2_init + (
            pos3_init - pos2_init
        ) / 2
    c.distances1 = c._get_distances(c.pos1, c.vec1, c.pos3)

    radius_temp1 = get_radius_critical(
        radius1,
        c.distances1,
        c.min_error
    )
    if radius_temp1 < c.radius_critical:
        c.radius_critical = radius_temp1
        assert c.radius_critical >= 0

    radius_effective1 = min(radius1, c.radius_critical)

    (
        c.tangent_length,
        c.dogleg
    ) = min_dist_to_target(
        radius_effective1,
        c.distances1
    )

    c.dogleg = check_dogleg(c.dogleg)

    c.dist_curve, c.func_dogleg = get_curve_hold_data(
                radius_effective1,
                c.dogleg
            )
    c.vec3 = get_vec_target(
        c.pos1,
        c.vec1,
        c.pos3,
        c.tangent_length,
        c.dist_curve,
        c.func_dogleg
    )

    tangent_temp1 = c._get_tangent_temp(c.tangent_length)

    c.pos2 = (
        c.pos3 - (
            tangent_temp1 * c.vec3
        )
    )

    c.distances2 = c._get_distances(
        c.pos_target,
        c.vec_target * -1,
        c.pos2
    )

    radius_temp2 = get_radius_critical(
        radius2,
        c.distances2,
        c.min_error
    )
    if radius_temp2 < c.radius_critical2:
        c.radius_critical2 = radius_temp2
        assert c.radius_critical2 >= 0

    radius_effective2 = min(radius2, c.radius_critical2)

    (
        c.tangent_length2,
        c.dogleg2
    ) = min_dist_to_target(
        radius_effective2,
        c.distances2
    )

    c.dogleg2 = check_dogleg(c.dogleg2)

    c.dist_curve2, c.func_dogleg2 = get_curve_hold_data(
                radius_effective2,
                c.dogleg2
            )
    c.vec2 = get_vec_target(
        c.pos_target,
        c.vec_target * -1,
        c.pos2,
        c.tangent_length2,
        c.dist_curve2,
        c.func_dogleg2
    )

    tangent_temp2 = c._get_tangent_temp(c.tangent_length2)

    c.pos3 = (
        c.pos2 - (
            tangent_temp2 * c.vec2
        )
    )

    vec23_denom = np.linalg.norm(c.pos3 - c.pos2)
    if vec23_denom == 0:
        c.vec23.append(np.array([0., 0., 0.]))
    else:
        c.vec23.append((c.pos3 - c.pos2) / vec23_denom)

    c.error = np.allclose(
        c.vec23[-1],
        vec_old,
        equal_nan=True,
        rtol=c.min_error * 10,
        atol=c.min_error * 0.1
    )

    c.errors.append(c.error)
    c.pos3_list.append(c.pos3)
    c.pos2_list.append(c.pos2)
    c.md_target = (
        c.md1 + c.dist_curve + tangent_temp2 + c.dist_curve2
    )
    c.delta_radius_list.append(
        abs(c.radius_critical - c.radius_critical2)
    )
    c.dls = max(
        dls_from_radius(c.radius_design),
        dls_from_radius(c.radius_critical)
    )
    c.dls2 = max(
        dls_from_radius(c.radius_design2),
        dls_from_radius(c.radius_critical2)
    )

    if c.error:
        if result:
            return c
        result = 0.
        if radius_temp2 < c.radius_design2:
            result += c.radius_design2 - radius_temp2
        if radius_temp1 < c.radius_design:
            result += c.radius_design - radius_temp1
        return result
    else:
        c.iterations += 1
        return minimize_target_pos_and_vec_defined(
            x, c, c.pos3, c.vec23[-1], result
        )


def check_dogleg(dogleg):
    # the code assumes angles are positive and clockwise
    if dogleg < 0:
        dogleg_new = dogleg + 2 * np.pi
        return dogleg_new
    else:
        return dogleg


def mod_vec(vec, error=1e-5):
    # if it's not working then twat it with a hammer
    vec_mod = vec * np.array([1, 1, 1 - error])
    vec_mod /= np.linalg.norm(vec_mod)
    inc_mod, azi_mod = get_angles(vec_mod, nev=True).T

    return vec_mod, inc_mod, azi_mod


def _get_xyz(pos):
        n, e, v = pos
        return np.array([e, n, v]).reshape(-1, 3)


def get_pos(pos1, vec1, vec2, dist_curve, func_dogleg):
    inc1, azi1 = _get_angles(_get_xyz(vec1)).reshape(2)
    inc2, azi2 = _get_angles(_get_xyz(vec2)).reshape(2)

    pos2 = pos1 + (
        (
            dist_curve * func_dogleg / 2
        )
        *
        np.array([
            np.sin(inc1) * np.cos(azi1) + np.sin(inc2) * np.cos(azi2),
            np.sin(inc1) * np.sin(azi1) + np.sin(inc2) * np.sin(azi2),
            np.cos(inc1) + np.cos(inc2)
        ]).T
    )

    return pos2


def get_vec_target(
    pos1,
    vec1,
    pos_target,
    tangent_length,
    dist_curve,
    func_dogleg
):
    if dist_curve == 0:
        return vec1

    vec_target = (
        (
            pos_target - pos1 - (
                    dist_curve
                    * func_dogleg
                ) / 2 * vec1
        )
        /
        (
            (
                dist_curve * func_dogleg / 2
            ) + tangent_length
        )
    )

    vec_target /= np.linalg.norm(vec_target)

    return vec_target


def get_curve_hold_data(radius, dogleg):
    dist_curve = radius * dogleg
    func_dogleg = shape_factor(dogleg)

    return (
        dist_curve,
        func_dogleg
    )


def shape_factor(dogleg):
    """
    Function to determine the shape factor of a dogleg.

    Parameters
    ----------
        dogleg: float
            The dogleg angle in radians of a curve section.
    """
    if dogleg == 0:
        return 0
    else:
        return np.tan(dogleg / 2) / (dogleg / 2)


def min_dist_to_target(radius, distances):
    """
    Calculates the control points for a curve and hold section from the
    start position and vector to the target position.
    """
    (
        dist_to_target,
        dist_perp_to_target,
        dist_norm_to_target
    ) = distances

    tangent_length = (
        dist_to_target ** 2
        - 2 * radius * dist_norm_to_target
    ) ** 0.5

    # determine the dogleg angle of the curve section
    dogleg = 2 * np.arctan2(
        (dist_perp_to_target - tangent_length),
        (
            2 * radius - dist_norm_to_target
        )
    )

    return tangent_length, dogleg


def min_curve_to_target(distances):
    """
    Calculates the control points for a curve section from the start
    position and vector to the target position which is not achievable with
    the provided dls_design.
    """
    if distances == (0., 0., 0,):
        return (
            0.,
            np.inf,
            0.
        )

    (
        dist_to_target,
        dist_perp_to_target,
        dist_norm_to_target
    ) = distances

    if dist_norm_to_target == 0.:
        radius_critical = np.inf
    else:
        radius_critical = (
            dist_to_target ** 2 / (
                2 * dist_norm_to_target
            )
        )
    assert radius_critical > 0

    dogleg = (
        2 * np.arctan2(
            dist_norm_to_target,
            dist_perp_to_target
        )
    )

    tangent_length = 0

    return (
        tangent_length,
        radius_critical,
        dogleg
    )


def get_radius_critical(radius, distances, min_error):
    (
        dist_to_target,
        dist_perp_to_target,
        dist_norm_to_target
    ) = distances

    if dist_norm_to_target == 0:
        return 0

    radius_critical = (
        dist_to_target ** 2 / (
            2 * dist_norm_to_target
        )
    ) * (1 - min_error)

    assert radius_critical > 0

    return radius_critical


def angle(vec1, vec2, acute=True):
    angle = np.arccos(
        np.dot(vec1, vec2)
        /
        (
            np.linalg.norm(vec1) * np.linalg.norm(vec2)
        )
    )

    if acute:
        return angle

    else:
        return 2 * np.pi - angle


def get_dogleg(inc1, azi1, inc2, azi2):
    dogleg = (
        2 * np.arcsin(
            (
                np.sin((inc2 - inc1) / 2) ** 2
                + np.sin(inc1) * np.sin(inc2)
                * np.sin((azi2 - azi1) / 2) ** 2
            ) ** 0.5
        )
    )

    return dogleg


def interpolate_well(sections, step=30):
    """
    Constructs a well survey from a list of sections of control points.

    Parameters
    ----------
    sections: list of welleng.connector.Connector objects
    step: float (default: 30)
        The desired delta measured depth between survey points, in
        addition to the control points.

    Returns
    -------
    data: list of welleng.connector.Connector objects
    """
    method = {
        'hold': get_interpolate_hold,
        'min_dist_to_target': get_interpolate_min_dist_to_target,
        'min_curve_to_target': get_interpolate_min_curve_to_target,
        'curve_hold_curve': get_interpololate_curve_hold_curve,
        'min_curve': get_min_curve
    }

    data = []
    if type(sections) is not list:
        sections = [sections]
    for s in sections:
        data.extend(method[s.method](s, step))

    return data


def interpolate_curve(
    md1,
    pos1,
    vec1,
    vec2,
    dist_curve,
    dogleg,
    func_dogleg,
    step,
    endpoint=False
):
    # sometimes the curve section has no length
    # this if statement handles this event
    if dist_curve == 0:
        inc, azi = get_angles(vec1, nev=True).T
        data = dict(
            md=np.array([md1]),
            vec=np.array([vec1]),
            inc=inc,
            azi=azi,
            dogleg=np.array([dogleg])
        )

        return data

    end_md = abs(dist_curve)
    if step is None:
        md = np.array([0])
    else:
        start_md = step - (md1 % step)
        md = np.arange(start_md, end_md, step)
        md = np.concatenate(([0.], md))
    if endpoint:
        md = np.concatenate((md, [end_md]))
    dogleg_interp = (dogleg / dist_curve * md).reshape(-1, 1)

    vec = (
        (
            np.sin(dogleg - dogleg_interp) / np.sin(dogleg) * vec1
        )
        +
        (
            np.sin(dogleg_interp) / np.sin(dogleg) * vec2
        )
    )
    vec = vec / np.linalg.norm(vec, axis=1).reshape(-1, 1)
    inc, azi = get_angles(vec, nev=True).T

    data = dict(
        md=md + md1,
        vec=vec,
        inc=inc,
        azi=azi,
        dogleg=np.concatenate((
            np.array([0.]), np.diff(dogleg_interp.reshape(-1))
        )),
    )

    return data


def interpolate_hold(md1, pos1, vec1, md2, step, endpoint=False):
    end_md = md2 - md1
    if step is None:
        md = np.array([0])
    else:
        start_md = step - (md1 % step)
        md = np.arange(start_md, end_md, step)
        md = np.concatenate(([0.], md))
    if endpoint:
        md = np.concatenate((md, [end_md]))
    vec = np.full((len(md), 3), vec1)
    inc, azi = get_angles(vec, nev=True).T
    dogleg = np.full_like(md, 0.)

    data = dict(
        md=md + md1,
        vec=vec,
        inc=inc,
        azi=azi,
        dogleg=dogleg,
    )

    return data


def get_min_curve(section, step=30, data=None):
    if section.md2 is None:
        result = (
            get_interpolate_min_curve_to_target(
                section, step, data
            )
        )
    else:
        result = (
            get_interpolate_min_dist_to_target(
                section, step, data
            )
        )
    return result


def get_interpolate_hold(section, step=30, data=None):
    if data is None:
        data = []

    data.append(interpolate_hold(
        md1=section.md1,
        pos1=section.pos1,
        vec1=section.vec1,
        md2=section.md_target,
        step=step,
        endpoint=True
    ))

    return data


def get_interpolate_min_curve_to_target(section, step=30, data=None):
    if data is None:
        data = []

    data.append(interpolate_curve(
        md1=section.md1,
        pos1=section.pos1,
        vec1=section.vec1,
        vec2=section.vec_target,
        dist_curve=section.dist_curve,
        dogleg=section.dogleg,
        func_dogleg=section.func_dogleg,
        step=step,
        endpoint=True
    ))

    return data


def get_interpolate_min_dist_to_target(section, step=30, data=None):
    if data is None:
        data = []

    # the first curve section
    data.append(interpolate_curve(
        md1=section.md1,
        pos1=section.pos1,
        vec1=section.vec1,
        vec2=section.vec2,
        dist_curve=section.dist_curve,
        dogleg=section.dogleg,
        func_dogleg=section.func_dogleg,
        step=step
    ))

    # the hold section
    data.append(interpolate_hold(
        md1=section.md2,
        pos1=section.pos2,
        vec1=section.vec2,
        md2=section.md_target,
        step=step,
        endpoint=True
    ))

    return data


def get_interpololate_curve_hold_curve(section, step=30, data=None):
    if data is None:
        data = []

    # the first curve section
    data.append(interpolate_curve(
        md1=section.md1,
        pos1=section.pos1,
        vec1=section.vec1,
        vec2=section.vec2,
        dist_curve=section.dist_curve,
        dogleg=section.dogleg,
        func_dogleg=section.func_dogleg,
        step=step
    ))

    # the hold section
    data.append(interpolate_hold(
        md1=section.md2,
        pos1=section.pos2,
        vec1=section.vec2,
        md2=section.md3,
        step=step
    ))

    # the second curve section
    data.append(interpolate_curve(
        md1=section.md3,
        pos1=section.pos3,
        vec1=section.vec3,
        vec2=section.vec_target,
        dist_curve=section.dist_curve2,
        dogleg=section.dogleg2,
        func_dogleg=section.func_dogleg2,
        step=step,
        endpoint=True
    ))

    return data


def convert_target_input_to_booleans(*inputs):
    input = [
        "0" if i is None else "1" for i in inputs
    ]

    return ''.join(input)


def connect_points(
    cartesians, vec_start=[0., 0., 1.], dls_design=3.0, nev=True,
    # step=30,
    md_start=0.
):
    """
    Function for connecting a list or array of only Cartesian points.

    Parameters
    ----------
        cartesians: (n, 3) list or array of floats
            Either [n, e, tvd] (default) or [x, y, z]
        vec_start: (3) list or array of floats (default: [0., 0., 1.])
            Unit start vector (default is pointing down) in the nev or xyz
            coordinate system.
        dls_design: float or (n, 1) list or array of floats (default: 3.0)
            The minimum Dog Leg Severity used when attempting to connect the
            points (a high DLS will be used if necessary).
        nev: bool (default: True)
            Indicates whether the cartesians are referencing the [nev]
            (default) or [xyz] coordinate system.
        step: float (default: 30)
            The desired step interval for the returned Survey object.
        md_start: float (default: 0)
            The md at the first cartesian point (in the event of a tie-on).

    Returns
    -------
        data: list of welleng.connector.Connector objects
    """
    if nev:
        pos_nev = np.array(cartesians).reshape(-1, 3)
        vec_nev = np.zeros_like(pos_nev)
        vec_nev[0] = np.array(vec_start).reshape(-1, 3)
    else:
        e, n, tvd = np.array(cartesians).reshape(-1, 3).T
        pos_nev = np.array([n, e, tvd]).T
        vec_nev = np.zeros_like(pos_nev)
        e, n, tvd = np.array(vec_start).reshape(-1, 3).T
        vec_nev[0] = np.array([n, e, tvd]).T

    if type(dls_design) is float:
        dls = np.full(len(pos_nev), dls_design)
    else:
        dls = np.array(dls_design).reshape(-1, 1)

    connections = []
    for i, (p, v, d) in enumerate(zip(pos_nev, vec_nev, dls)):
        if i == 0:
            node_1 = Node(
                pos=p,
                vec=v,
                md=md_start
            )
            continue
        if i > 1:
            node_1 = connections[-1].node_end
        node_2 = Node(
            pos=p
        )
        c = Connector(
            node1=node_1,
            node2=node_2,
            dls_design=d
        )
        assert np.allclose(c.pos_target, p)
        connections.append(c)

    return connections


def survey_to_plan(survey, tolerance=0.2, dls_design=1., step=30.):
    """
    Prototype function for extracting a plan from a drilled well survey - a
    minimal number of control points (begining/end points of either hold or
    build/turn sections) required to express the well given the provided input
    parameters.

    Parameters
    ----------
    survey: welleng.survey.Survey object
    tolerance: float (default=0.2)
        Defines how tight to fit the planned well to the survey - a higher
        number will results in less control points but likely poorer fit.
    dls_design: float (default=1.0)
        The minimum DLS used to fit the planned trajectory.
    step: float (default=30)
        The desired md step in the plan survey.

    Returns
    -------
    sections: list of welleng.connector.Connection objects
    """
    assert dls_design > 0., "dls_design must be greater than 0"

    idx = [0]
    end = len(survey.md) - 1
    md = survey.md[0]
    node = None
    sections = []

    while True:
        section, i = _get_section(
            survey=survey,
            start=idx[-1],
            md=md,
            node=node,
            tolerance=tolerance,
            dls_design=dls_design
        )
        sections.append(section)
        idx.append(i)
        if idx[-1] >= end:
            break
        node = section.node_end

    data = interpolate_well(sections, step=30)

    return sections


def _get_section(
    survey, start, tolerance, dls_design=1., md=0., node=None
):
    idx = start + 2
    nev = survey.get_nev_arr()

    if idx > len(nev):
        idx = len(nev) - 1

    if node is None:
        node_1 = Node(
            pos=nev[start],
            vec=survey.vec_nev[start],
            md=md
        )
    else:
        node_1 = node

    scores = [0.]
    delta_scores = []
    c_old = None

    for i, (p, v) in enumerate(zip(nev[idx:], survey.vec_nev[idx:])):
        node_2 = Node(
            pos=p,
            vec=v,
        )
        c = Connector(
            node1=node_1,
            node2=node_2,
            dls_design=dls_design
        )
        s = c.survey(step=1.)
        if c_old is None:
            c_old = deepcopy(c)
        nev_new = s.get_nev_arr()

        distances = distance.cdist(
            nev[start: idx + i],
            nev_new
        )

        score = np.sum(np.amin(distances, axis=1)) / (s.md[-1] - s.md[0])

        delta_scores.append(score - scores[-1])
        scores.append(score)

        if all((
            abs(delta_scores[-1]) >= tolerance,
            idx + i < len(nev) - 2
        )):
            break
        elif idx + i == len(nev) - 1:
            c_old = deepcopy(c)
            break
        else:
            c_old = deepcopy(c)

    return (
        c_old,
        idx + i - 1 if idx + i != len(nev) - 1 else idx + i
    )


def numbafy(functions):
    for f in functions:
        f = njit(f)


if NUMBA:
    NUMBAFY = (
        _get_xyz,
        get_pos,
        get_vec_target,
        get_curve_hold_data,
        shape_factor,
        min_dist_to_target,
        get_radius_critical,
        angle,
        get_dogleg
    )
    numbafy(NUMBAFY)



import os
import numpy as np
import yaml
from .errors.tool_errors import ToolError

# TODO: there's likely an issue with TVD versus TVDSS that
# needs to be resolved. This model assumes TVD relative to
# rig floor, but often a TVDSS is provided instead (with a
# negative value for rig floor elevation).

ACCURACY = 1e-4
PATH = os.path.dirname(__file__)
TOOL_INDEX_FILENAME = os.path.join(
    '', *[PATH, 'errors', 'tool_index.yaml']
)


def get_tool_index():
    with open(TOOL_INDEX_FILENAME, 'r') as f:
        tool_index = yaml.safe_load(f)
    return tool_index


def get_error_models(tool_index=None):
    if tool_index is None:
        tool_index = get_tool_index()
    error_models = [
        v['Short Name']
        for _, v in tool_index.items()
    ]
    return error_models


TOOL_INDEX = get_tool_index()
ERROR_MODELS = get_error_models(TOOL_INDEX)


class ErrorModel():
    """
    A class to initiate the field parameters and error magnitudes
    for subsequent error calculations.
    """

    class Error:
        '''
        Standard components of a well bore survey error.
        '''
        def __init__(
            self,
            code,
            propagation,
            e_DIA,
            cov_DIA,
            e_NEV,
            e_NEV_star,
            sigma_e_NEV,
            cov_NEV
        ):

            self.code = code
            self.propagation = propagation
            self.e_DIA = e_DIA
            self.cov_DIA = cov_DIA
            self.e_NEV = e_NEV
            self.e_NEV_star = e_NEV_star
            self.sigma_e_NEV = sigma_e_NEV
            self.cov_NEV = cov_NEV

    def __init__(
        self,
        survey,
        error_model="ISCWSA MWD Rev5",
    ):
        """

        """

        # error_models = ERROR_MODELS
        assert error_model in ERROR_MODELS, "Unrecognized error model"
        self.error_model = error_model
        self.survey = survey

        self.survey_rad = np.stack((
            self.survey.md,
            self.survey.inc_rad,
            self.survey.azi_true_rad
        ), axis=-1)

        self.survey_drdp = self.survey_rad
        self.drdp = self._drdp(self.survey_drdp)
        self.drdp_sing = self._drdp_sing(self.survey_drdp)

        # if self.error_model.split("_")[0] == "iscwsa":
        #     self.errors = iscwsaMwd(
        #         error=self,
        #         model=self.error_model
        #     )

        for k, v in TOOL_INDEX.items():
            if v['Short Name'] == self.error_model:
                model = k
                break

        self.errors = ToolError(
            error=self,
            model=model
        )

    def _e_NEV(self, e_DIA):
        D, I, A = e_DIA.T
        arr = np.array([
            (self.drdp[:, 0] + self.drdp[:, 9]) * D
            + (self.drdp[:, 3] + self.drdp[:, 12]) * I
            + (self.drdp[:, 6] + self.drdp[:, 15]) * A,

            (self.drdp[:, 1] + self.drdp[:, 10]) * D
            + (self.drdp[:, 4] + self.drdp[:, 13]) * I
            + (self.drdp[:, 7] + self.drdp[:, 16]) * A,

            (self.drdp[:, 2] + self.drdp[:, 11]) * D
            + (self.drdp[:, 5] + self.drdp[:, 14]) * I
            + (self.drdp[:, 8] + self.drdp[:, 17]) * A,
        ]).T

        arr[0] = 0

        return arr

    def _e_NEV_star(self, e_DIA):
        D, I, A = e_DIA.T
        arr = np.array([
            self.drdp[:, 0] * D
            + self.drdp[:, 3] * I
            + self.drdp[:, 6] * A,

            self.drdp[:, 1] * D
            + self.drdp[:, 4] * I
            + self.drdp[:, 7] * A,

            self.drdp[:, 2] * D
            + self.drdp[:, 5] * I
            + self.drdp[:, 8] * A
        ]).T

        arr[0] = 0

        return arr

    def _cov(self, arr):
        '''
        Returns a covariance matrix from an (n,3) array.
        '''
        # Mitigate overflow
        # with np.errstate(divide='ignore', invalid='ignore'):
        #     coeff = np.nan_to_num(
        #         arr / np.abs(arr) * ACCURACY,
        #         nan=ACCURACY
        #     )
        # arr = np.where(np.abs(arr) > ACCURACY, arr, coeff)

        x, y, z = np.array(arr).T
        result = np.array([
            [x*x, x*y, x*z],
            [y*x, y*y, y*z],
            [z*x, z*y, z*z]
        ])

        return result

    def _sigma_e_NEV_systematic(self, e_NEV, e_NEV_star):
        return e_NEV_star + np.vstack(
            (
                e_NEV[0],
                np.cumsum(e_NEV, axis=0)[:-1]
            )
        )

    def _generate_error(
        self,
        code,
        e_DIA,
        propagation='systematic',
        NEV=True,
        e_NEV=None,
        e_NEV_star=None
    ):
        if not NEV:
            return e_DIA
        else:
            cov_DIA = self._cov(e_DIA)
            if e_NEV is None:
                e_NEV = self._e_NEV(e_DIA)
                e_NEV_star = self._e_NEV_star(e_DIA)
            if propagation == 'systematic':
                sigma_e_NEV = self._sigma_e_NEV_systematic(e_NEV, e_NEV_star)
                cov_NEV = self._cov(sigma_e_NEV)
            elif propagation == 'random':
                sigma_e_NEV = np.cumsum(self._cov(e_NEV), axis=-1)
                cov_NEV = np.add(
                    self._cov(e_NEV_star),
                    np.concatenate(
                        (
                            np.array(np.zeros((3, 3, 1))),
                            np.array(sigma_e_NEV[:, :, :-1])
                        ), axis=-1)
                    )
            else:
                return

            return ErrorModel.Error(
                code,
                propagation,
                e_DIA,
                cov_DIA,
                e_NEV,
                e_NEV_star,
                sigma_e_NEV,
                cov_NEV
            )

    def drk_dDepth(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        # TODO: This is essentially minimum curvature... use function from
        # utils instead (it's already in self.mc)
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T

        N = np.array(
            0.5 * (
                np.sin(inc1) * np.cos(azi1)
                + np.sin(inc2) * np.cos(azi2)
            )
        )

        E = np.array(
            0.5 * (
                np.sin(inc1) * np.sin(azi1)
                + np.sin(inc2) * np.sin(azi2)
            )
        )

        V = np.array(
            0.5 * (
                np.cos(inc1) + np.cos(inc2)
            )
        )

        return np.vstack(
            (
                np.array(np.zeros((1, 3))),
                np.stack((N, E, V), axis=-1)
            )
        )

    def drk_dInc(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T
        delta_md = md2 - md1

        N = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.cos(azi2)))
        E = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.sin(azi2)))
        V = np.array(0.5 * (-delta_md * np.sin(inc2)))

        if self.error_model.lower().split()[-1] != 'rev4':
            N[0] *= 2

        return np.vstack(
            (
                np.array(np.zeros((1, 3))),
                np.stack((N, E, V), axis=-1)
            )
        )

    def drk_dAz(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T
        delta_md = md2 - md1

        N = np.array(-0.5 * ((delta_md) * np.sin(inc2) * np.sin(azi2)))
        E = np.array(0.5 * ((delta_md) * np.sin(inc2) * np.cos(azi2)))
        V = np.zeros_like(N)

        return np.vstack(
            (
                np.array(np.zeros((1, 3))),
                np.stack((N, E, V), axis=-1)
            )
        )

    def drkplus1_dDepth(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''
        return np.vstack(
            (
                self.drk_dDepth(survey)[1:] * -1,
                np.array(np.zeros((1, 3)))
            )
        )

    def drkplus1_dInc(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''

        md2, inc2, azi2 = np.array(survey[:-1]).T
        md3, inc3, azi3 = np.array(survey[1:]).T
        delta_md = md3 - md2

        N = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.cos(azi2)))
        E = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.sin(azi2)))
        V = np.array(0.5 * (-(delta_md) * np.sin(inc2)))

        return np.vstack(
            (
                np.stack((N, E, V), axis=-1),
                np.array(np.zeros((1, 3)))
            )
        )

    def drkplus1_dAz(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''
        md2, inc2, azi2 = np.array(survey[:-1]).T
        md3, inc3, azi3 = np.array(survey[1:]).T
        delta_md = md3 - md2

        N = np.array(-0.5 * ((delta_md) * np.sin(inc2) * np.sin(azi2)))
        E = np.array(0.5 * ((delta_md) * np.sin(inc2) * np.cos(azi2)))
        V = np.zeros_like(N)

        return np.vstack(
            (
                np.stack((N, E, V), axis=-1),
                np.array(np.zeros((1, 3)))
            )
        )

    def _drdp(self, survey):

        return np.hstack((
            self.drk_dDepth(survey),
            self.drk_dInc(survey),
            self.drk_dAz(survey),
            self.drkplus1_dDepth(survey),
            self.drkplus1_dInc(survey),
            self.drkplus1_dAz(survey)
        ))

    def _drdp_sing(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station
        survey3 is next survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-2]).T
        md2, inc2, azi2 = np.array(survey[1:-1]).T
        md3, inc3, azi3 = np.array(survey[2:]).T
        double_delta_md = md3 - md1
        delta_md = md2 - md1

        return dict(
            double_delta_md=double_delta_md,
            delta_md=delta_md,
            azi2=azi2
        )


def get_errors(error):
    nn, ne, nv = error[0]
    _, ee, ev = error[1]
    _, __, vv = error[2]

    return [nn, ee, vv, ne, nv, ev]


def make_diagnostic_data(survey):
    diagnostic = {}
    dia = np.stack((survey.md, survey.inc_deg, survey.azi_grid_deg), axis=1)
    for i, d in enumerate(survey.md):
        diagnostic[d] = {}
        total = []
        for k, v in survey.err.errors.errors.items():
            diagnostic[d][k] = get_errors(v.cov_NEV.T[i])
            total.extend(diagnostic[d][k])
        diagnostic[d]['TOTAL'] = np.sum((np.array(
            total
        ).reshape(-1, len(diagnostic[d][k]))), axis=0)
    return diagnostic



try:
    import trimesh
    TRIMESH = True
except ImportError:
    TRIMESH = False
try:
    from vedo import trimesh2vedo, Lines, Sphere
    VEDO = True
except ImportError:
    VEDO = False
import numpy as np

try:
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    PLOTLY = True
except ImportError:
    PLOTLY = False

from vtk import (
        vtkCubeAxesActor, vtkNamedColors, vtkInteractorStyleTerrain
)
from vtkmodules.vtkRenderingCore import (
    vtkRenderWindow, vtkRenderWindowInteractor, vtkRenderer
)

from .version import __version__ as VERSION

# VEDO = False

class Plotter(vtkRenderer):
    def __init__(self, data=None, **kwargs):
        super().__init__()

        """
        A vtk wrapper for quickly visualizing well trajectories for QAQC
        purposes. Initiates a vtkRenderer instance and populates it with
        mesh data and a suitable axes for viewing well trajectory data.

        Parameters
        ----------
        data: obj or list(obj)
            A trimesh.Trimesh object or a list of trimesh.Trimesh
            objects or a trimesh.scene object
        names: list of strings (default: None)
            A list of names, index aligned to the list of well meshes.
        colors: list of strings (default: None)
            A list of color or colors. If a single color is listed then this is
            applied to all meshes in data, otherwise the list of colors is
            indexed to the list of meshes.
        """
        assert all((VEDO, TRIMESH)), \
            "ImportError: try pip install welleng[easy]"

        if data is not None:
            names = kwargs.get('names')

            if isinstance(data, trimesh.scene.scene.Scene):
                meshes = [v for k, v in data.geometry.items()]
                if names is None:
                    names = list(data.geometry.keys())

            # handle a single mesh being passed
            elif isinstance(data, trimesh.Trimesh):
                meshes = [data]

            else:
                meshes = data
                if names is not None:
                    assert len(names) == len(data), \
                        "Names must be length of meshes list else None"

            colors = kwargs.get('colors')
            if colors is not None:
                if len(colors) == 1:
                    colors = colors * len(meshes)
                else:
                    assert len(colors) == len(meshes), \
                        "Colors must be length of meshes list, 1 else None"

            points = kwargs.get('points')
            if points is not None:
                points = [
                    Sphere(p, r=30, c='grey')
                    for p in points
                ]

            meshes_vedo = []
            for i, mesh in enumerate(meshes):
                if i == 0:
                    vertices = np.array(mesh.vertices)
                    start_locations = np.array([mesh.vertices[0]])
                else:
                    vertices = np.concatenate(
                        (vertices, np.array(mesh.vertices)),
                        axis=0
                    )
                    start_locations = np.concatenate(
                        (start_locations, np.array([mesh.vertices[0]])),
                        axis=0
                    )

                # convert to vedo mesh
                m_vedo = trimesh2vedo(mesh)
                if colors is not None:
                    m_vedo.c(colors[i])
                if names is not None:
                    m_vedo.name = names[i]
                    m_vedo.flag()
                meshes_vedo.append(m_vedo)

                for mesh in meshes_vedo:
                    if mesh is None:
                        continue
                    self.AddActor(mesh)

                for obj in kwargs.values():
                    if isinstance(obj, list):
                        for item in obj:
                            try:
                                self.AddActor(item)
                            except TypeError:
                                pass
                    else:
                        try:
                            self.AddActor(obj)
                        except TypeError:
                            pass

        self.namedColors = vtkNamedColors()

        self.colors = {}
        self.colors['background'] = kwargs.get('background', 'LightGrey')
        self.colors['background2'] = kwargs.get('background', 'Lavender')

    def add_axes(self, **kwargs):
        axes = CubeAxes(self, **kwargs)
        self.AddActor(axes)
    
    def get_center(self):
        min, max = np.array(self.ComputeVisiblePropBounds()).reshape(-1, 3)
        center = min + (max - min) / 2

        return tuple(center)

    def show(self, add_axes=True, **kwargs):
        """
        Convenient method for opening a window to view the rendered scene.
        """
        if add_axes:
            self.add_axes(**kwargs)
        
        self.GetActiveCamera().Azimuth(kwargs.get('azimuth', 30))
        self.GetActiveCamera().Elevation(kwargs.get('elevation', 30))
        self.GetActiveCamera().SetViewUp(0, 0, -1)
        # self.GetActiveCamera().SetPosition(tuple(pos_new))
        self.GetActiveCamera().SetFocalPoint(self.get_center())

        self.ResetCamera()
        self.SetBackground(
            self.namedColors.GetColor3d(self.colors['background'])
        )
        self.SetBackground2(
            self.namedColors.GetColor3d(self.colors['background2'])
        )

        setSize = kwargs.get('setSize', (1200, 900))

        renderWindow = vtkRenderWindow()

        renderWindow.AddRenderer(self)
        renderWindow.SetSize(*(setSize))
        renderWindow.SetWindowName(f'welleng {VERSION}')

        renderWindowInteractor = vtkRenderWindowInteractor()
        renderWindowInteractor.SetRenderWindow(renderWindow)
        renderWindowInteractor.SetInteractorStyle(vtkInteractorStyleTerrain())

        renderWindow.Render()
        self.GetActiveCamera().Zoom(0.8)

        interactive = kwargs.get('interactive', True)
        if interactive:
            renderWindowInteractor.Start()


class CubeAxes(vtkCubeAxesActor):
    def __init__(self, renderer, **kwargs):
        super().__init__()

        # Determine the bounds from the meshes/actors being plotted rounded
        # up to the nearest 100 units.
        bounds = np.array(renderer.ComputeVisiblePropBounds())

        with np.errstate(divide='ignore', invalid='ignore'):
            self.bounds = tuple(np.nan_to_num(
                (
                    np.ceil(np.abs(bounds) / 100) * 100
                ) * (bounds / np.abs(bounds))
            ))

        namedColors = vtkNamedColors()
        self.colors = {}

        for n in range(1, 4):
            self.colors[f'axis{n}Color'] = namedColors.GetColor3d(
                kwargs.get('axis1Color', 'DarkGrey')
            )

        self.SetLabelScaling(0, 0, 0, 0)

        # Try and prevent scientific numbering on axes
        self.SetXLabelFormat("%.0f")
        self.SetYLabelFormat("%.0f")
        self.SetZLabelFormat("%.0f")

        self.SetUseTextActor3D(1)

        self.SetBounds(self.bounds)
        self.SetCamera(renderer.GetActiveCamera())

        self.GetTitleTextProperty(0).SetColor(self.colors['axis1Color'])
        self.GetLabelTextProperty(0).SetColor(self.colors['axis1Color'])
        self.GetLabelTextProperty(0).SetOrientation(45.0)

        self.GetTitleTextProperty(1).SetColor(self.colors['axis2Color'])
        self.GetLabelTextProperty(1).SetColor(self.colors['axis2Color'])
        self.GetLabelTextProperty(1).SetOrientation(45.0)

        self.GetTitleTextProperty(2).SetColor(self.colors['axis3Color'])
        self.GetLabelTextProperty(2).SetColor(self.colors['axis3Color'])
        self.GetLabelTextProperty(2).SetOrientation(45.0)

        self.DrawXGridlinesOn()
        self.DrawYGridlinesOn()
        self.DrawZGridlinesOn()
        self.SetGridLineLocation(self.VTK_GRID_LINES_FURTHEST)

        self.XAxisMinorTickVisibilityOff()
        self.YAxisMinorTickVisibilityOff()
        self.ZAxisMinorTickVisibilityOff()

        units = kwargs.get('units', 'meters')
        self.SetXTitle('East')
        self.SetXUnits(units)
        self.SetYTitle('North')
        self.SetYUnits(units)
        self.SetZTitle('TVD')
        self.SetZUnits(units)

        self.SetFlyModeToClosestTriad()

        self.ForceOpaqueOff()


def plot(
    data,
    names=None,
    colors=None,
    lines=None,
    targets=None,
    arrows=None,
    text=None,
    boxes=None,
    points=None,
    **kwargs
):
    """
    A deprecated wrapper for the Plotter class, maintained only for
    compatability with older versions.

    Parameters
    ----------
    data: a trimesh.Trimesh object or a list of trimesh.Trimesh
    objects or a trmiesh.scene object
    names: list of strings (default: None)
        A list of names, index aligned to the list of well meshes.
    colors: list of strings (default: None)
        A list of color or colors. If a single color is listed then this is
        applied to all meshes in data, otherwise the list of colors is
        indexed to the list of meshes.
    """
    for k, v in locals().items():
        if k != 'data' and k[0] != '_':
            kwargs.update({k: v})

    plt = Plotter(data, **kwargs)

    plt.show()


def get_lines(clearance):
    """
    Add lines per reference well interval between the closest points on the
    reference well and the offset well and color them according to the
    calculated Separation Factor (SF) between the two wells at these points.

    Parameters
    ----------
        clearance: welleng.clearance object

    Returns
    -------
        lines: vedo.Lines object
            A vedo.Lines object colored by the object's SF values.
    """
    assert VEDO, "ImportError: try pip install welleng[easy]"
    c = clearance.SF
    start_points, end_points = clearance.get_lines()
    lines = Lines(start_points, end_points).cmap('hot_r', c, on='cells')
    lines.addScalarBar(title='SF')

    return lines


def figure(obj, type='scatter3d', **kwargs):
    assert PLOTLY, "ImportError: try pip install plotly"
    func = {
        'scatter3d': _scatter3d,
        'mesh3d': _mesh3d,
        'panel': _panel
    }
    fig = func[type](obj, **kwargs)
    return fig


def _panel(survey, **kwargs):
    fig = make_subplots(
        rows=2, cols=2,
        subplot_titles=(
            "Plan",
            f"Vertical Section: {np.degrees(survey.header.vertical_section_azimuth)} deg",
            "WE Section", "NS Section"
        )
        # shared_xaxes=True, shared_yaxes=True
    )

    fig.add_trace(
        go.Scatter(
            x=survey.e,
            y=survey.n,
            mode='lines',
            name='NE Plan',
            showlegend=False
        ),
        row=1, col=1
    )
    fig.add_trace(
        go.Scatter(
            x=survey.vertical_section,
            y=survey.tvd,
            mode='lines',
            name='Plan',
            showlegend=False
        ),
        row=1, col=2
    )
    fig.add_trace(
        go.Scatter(
            x=survey.n,
            y=survey.tvd,
            mode='lines',
            name='NS Section',
            showlegend=False,
        ),
        row=2, col=2
    )
    fig.add_trace(
        go.Scatter(
            x=survey.e,
            y=survey.tvd,
            mode='lines',
            name='WE Section',
            showlegend=False
        ),
        row=2, col=1
    )

    fig.update_layout(
        xaxis=dict(
            title='West-East'
        ),
        yaxis=dict(
            title='North-South'
        ),
        xaxis2=dict(
            title='Outstep'
        ),
        yaxis2=dict(
            title='TVD',
            autorange="reversed",
            matches='y3'
        ),
        xaxis3=dict(
            title='West-East',
            matches='x'
        ),
        yaxis3=dict(
            title='TVD',
            autorange="reversed"
        ),
        xaxis4=dict(
            title='North-South',
            matches='y'
        ),
        yaxis4=dict(
            title='TVD',
            autorange="reversed",
            matches='y3'
        ),
    )

    return fig


def _scatter3d(survey, **kwargs):
    fig = go.Figure()
    fig.add_trace(
        go.Scatter3d(
            x=survey.e,
            y=survey.n,
            z=survey.tvd,
            name='survey',
            mode='lines',
            hoverinfo='skip'
        )
    )
    if not hasattr(survey, "interpolated"):
        survey.interpolated = [False] * len(survey.md)
    if survey.interpolated is None:
        survey.interpolated = [False] * len(survey.md)
    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is True
        ]).T
        if n.size:
            text = [
                f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
                + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
                for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
            ]
            fig.add_trace(
                go.Scatter3d(
                    x=e,
                    y=n,
                    z=v,
                    name='interpolated',
                    mode='markers',
                    marker=dict(
                        size=5,
                        color='blue',
                    ),
                    text=text,
                    hoverinfo='text'
                )
            )
    except ValueError:
        pass

    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is False
        ]).T
        text = [
            f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
            + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
            for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
        ]
        fig.add_trace(
            go.Scatter3d(
                x=e,
                y=n,
                z=v,
                name='survey_point',
                mode='markers',
                marker=dict(
                    size=5,
                    color='red',
                ),
                text=text,
                hoverinfo='text'
            )
        )
    except ValueError:
        pass

    fig = _update_fig(fig, kwargs)

    return fig


def _mesh3d(mesh, **kwargs):
    fig = go.Figure()
    vertices = mesh.vertices.reshape(-1, 3)
    faces = np.array(mesh.faces)
    n, e, v = vertices.T
    i, j, k = faces.T
    fig.add_trace(
        go.Mesh3d(
            x=e, y=n, z=v,
            i=i, j=j, k=k,
        )
    )
    if kwargs.get('edges'):
        tri_points = np.array([
            vertices[i] for i in faces.reshape(-1)
        ])
        n, e, v = tri_points.T
        fig.add_trace(
            go.Scatter3d(
                x=e, y=n, z=v,
                mode='lines',
            )
        )
    fig = _update_fig(fig, kwargs)

    return fig


def _update_fig(fig, kwargs):
    """
    Update the fig axis along with any user defined kwargs.
    """
    fig.update_scenes(
        zaxis_autorange="reversed",
        aspectmode='data',
        xaxis=dict(
            title='East (m)'
        ),
        yaxis=dict(
            title='North (m)',
        ),
        zaxis=dict(
            title="TVD (m)"
        )
    )
    for k, v in kwargs.items():
        if k == "layout":
            fig.update_layout(v)
        elif k == "traces":
            fig.update_traces(v)
        else:
            continue

    return fig



try:
    from openpyxl import load_workbook
    OPENPYXL = True
except:
    OPENPYXL = False
from math import radians
import yaml
import os

"""
Code to extract error model data from standard ISCWSA Excel files and create
yaml files for use with the welleng.error module.

It's coded up this way so that when the error models are updated then simply
add the new workbook to the reference folder and add a new line to remake
the error model input data.

If new error functions are added then these will need to be coded into the
error module.
"""

PATH = os.path.dirname(__file__)

FILENAME = 'welleng/errors/error_codes.yaml'
CHARACTERS = [":", "[", "]"]

try:
    with open(FILENAME, 'r') as file:
        ec = yaml.full_load(file)
except Exception:
    ec = {}


def extract_data(filename):
    assert OPENPYXL, "ImportError: try pip install welleng[easy]"
    workbook = load_workbook(
        filename=filename,
        data_only=True
    )
    m = workbook['Model']

    e = extract_codes(m)
    h = extract_header(m)

    ed = dict(
        header=h,
        codes=e
    )

    return ed


def extract_header(workbook):
    m = workbook
    d = []

    for row in m.iter_rows(
        min_row=0,
        max_row=(),
        min_col=0,
        max_col=2,
        values_only=True
    ):
        if row[0] is None:
            continue
        else:
            d.append([
                row[0], row[1]
            ])

    h = make_header_dict(d)

    return h


def extract_codes(workbook):
    m = workbook
    d = []

    for row in m.iter_rows(
        min_row=4,
        max_row=(),
        min_col=4,
        max_col=(),
        values_only=True
    ):
        code = row[1]
        if code == "XCLL":
            code = "XCLA"
        function = row[3]
        magnitude = row[6]
        unit = row[7]
        propagation = row[8]
        d.append([
            code, function, magnitude, unit, propagation
        ])

    e = make_error_dict(d)

    return e


def make_error_dict(data):
    e = {}

    for c, f, m, u, p in data:
        e[c] = dict(
            function=f.replace('-', '_'),
            magnitude=m if 'deg' not in u else radians(m),
            unit=u.replace("deg", "rad"),
            propagation='random' if p == 'R' else 'systematic'
        )

    return e


def make_header_dict(data):
    h = {}

    for k, v in data:
        kk = remove_characters(k)
        h[kk] = v

    return h


def remove_characters(data, chars=CHARACTERS):
    for c in chars:
        data = data.replace(c, "")

    return data


def make_index(tool_code, data):
    fields = [
        'Application', 'Short Name', 'Long Name', 'OWSG Prefix', 'Revision No'
    ]
    temp = {
        f: data['header'][f] for f in fields
    }
    temp = {
        (k if k != 'Revision No' else 'Rev.'): v
        for k, v in temp.items()
    }
    filename = os.path.join(
        '',
        *[PATH, 'tool_index.yaml']
    )

    # read tool_index file
    with open(filename, 'r') as f:
        tool_index = yaml.safe_load(f)

    # add (or overwrite) new entry
    tool_index[tool_code] = temp

    # save new tool_index
    with open(filename, 'w') as f:
        yaml.dump(tool_index, f)


if __name__ == '__main__':
    # # extract rev 4 model
    tool_code = 'ISCWSA MWD Rev4'
    ec = extract_data(
        "reference/error-model-example-mwdrev4-iscwsa-1.xlsx"
    )

    # write data to yaml
    filename = os.path.join(
        '',
        *[PATH, 'tool_codes', f'{tool_code}.yaml']
    )

    with open(filename, 'w') as f:
        documents = yaml.dump(ec, f)

    # update index file
    make_index(tool_code, ec)


    # extract rev 5 model
    tool_code = 'ISCWSA MWD Rev5'
    ec = extract_data(
        "reference/error-model-example-mwdrev5-iscwsa-1.xlsx"
    )

    filename = os.path.join(
        '',
        *[PATH, 'tool_codes', f'{tool_code}.yaml']
    )

    with open(filename, 'w') as f:
        documents = yaml.dump(ec, f)

    make_index(tool_code, ec)

    print("Done")



import numbers
import numpy as np
from .survey import Survey

try:
    from openpyxl import load_workbook
    OPENPYXL = True
except ImportError:
    OPENPYXL = False


def get_standard_data(filename):
    # import data from Excel
    assert OPENPYXL, "ImportError: try pip install welleng[easy]"
    workbook = load_workbook(filename, data_only=True)
    sheets = workbook.sheetnames

    # extract the sheets with the input data
    wells = []
    wells.extend([well for well in list(sheets) if well.split()[-1] == "well"])

    data = {
        "acr": {},
        "wells": {},
    }

    for i, well in enumerate(wells):
        sheet = workbook[well]
        data = get_well_data(well, sheet, data)
        if well == "Reference well":
            data = acr_setup(sheet, data)
        else:
            sheet = workbook[f"{well[:5]}clearance"]
            data = get_clearance_data(well, sheet, data)

    return data


def get_well_data(well, sheet, data):
    temp = dict(
        MD=[],
        IncDeg=[],
        AziDeg=[],
        TVD=[],
        N=[],
        E=[],
        sigmaH=[],
        sigmaL=[],
        sigmaA=[],
    )

    for row in sheet.iter_rows(
        min_row=17,
        max_row=(),
        min_col=2,
        max_col=10,
    ):
        if isinstance(row[0].value, numbers.Real):
            temp["MD"].append(row[0].value)
            temp["IncDeg"].append(row[1].value)
            temp["AziDeg"].append(row[2].value)
            temp["TVD"].append(row[3].value)
            temp["N"].append(row[4].value)
            temp["E"].append(row[5].value)
            temp["sigmaH"].append(row[6].value)
            temp["sigmaL"].append(row[7].value)
            temp["sigmaA"].append(row[8].value)

    data["wells"][f"{well}"] = temp

    return data


def acr_setup(sheet, data):
    data["acr"]["Sm"] = sheet["I4"].value
    data["acr"]["sigmapa"] = sheet["I5"].value
    data["acr"]["k"] = sheet["I6"].value
    data["acr"]["reference_h_and_c"] = sheet["I7"].value
    data["acr"]["offset_h_and_c"] = sheet["I8"].value

    return data


def make_survey(data, well):
    start_nev = data["wells"]["offset"]["mesh"].NEVs[0]
    y, x, z = start_nev
    start_xyz = np.array([x, y, z])
    return Survey(
        md=data["wells"][well]["MD"],
        inc=data["wells"][well]["IncDeg"],
        azi=data["wells"][well]["AziDeg"],
        start_nev=start_nev,
        start_xyz=start_xyz
    )


def get_clearance_data(well, sheet, data):
    sf = []
    for row in sheet.iter_rows(
        min_row=5,
        max_row=(),
        min_col=15,
        max_col=16,
    ):
        sf.append(row[0].value)

    data["wells"][f"{well}"]['SF'] = sf

    return data


def import_iscwsa_collision_data(filename):
    data = get_standard_data(filename)

    return data


if __name__ == "__main__":
    filename = (
        "reference/standard-set-of-wellpaths"
        "-for-evaluating-clearance-scenarios-r4-17-may-2017.xlsm"
    )
    data = import_iscwsa_collision_data(filename)



PARAMS = {
    'WellBore': ['id', 'coeff_friction_sliding']
}

class String:
    def __init__(self, name, top, bottom, *args, method="bottom_up", **kwargs):
        """
        A generic well bore architecture collection, e.g. a casing string
        made up a a number of different lengths of weights and grades.

        Parameters
        ----------
        name: str
            The name of the collection.
        top: float
            The shallowest measured depth at the top of the collection of
            items in meters.
        bottom: float
            The deepest measured depth at the bottom of the collection of
            items in meters.
        method: string (default: 'bottom up')
            The method in which items are added to the collection, either
            'bottom up' starting from the deepest element and adding items
            above, else 'top down' starting from the shallowest item and
            adding items below.
        """
        self.name = name
        self.top = top
        self.bottom = bottom
        self.sections = {}
        self.complete = False

        assert method in {"top_down", "bottom_up"}, "Unrecognized method"
        self.method = method

    def depth(self, md):
        assert self.top < md <= self.bottom, "Depth out of range"

        string_new = String(
            self.name, self.top, md, method="bottom_up"
        )

        reached_top = False
        for section in reversed(list(self.sections.keys())):
            if reached_top:
                break
            params = {
                    k: v for k, v in self.sections[section].items()
                    if k not in ['top', 'bottom', 'length', 'buoyancy_factor']
            }
            string_new.add_section(
                length=self.sections[section]['length'], **params
            )
            if string_new.sections[0]['top'] == self.top:
                reached_top = True

        return string_new

    def add_section(self, **kwargs):
        if type(self).__name__ == 'WellBore':
            for param in PARAMS.get('WellBore'):
                assert param in kwargs.keys(), f"Missing parameter {param}"

        elif type(self).__name__ == 'BHA':
            kwargs['density'] = kwargs.get('density', 7.85)

        if self.method == "top_down":
            self.add_section_top_down(**kwargs)
        elif self.method == "bottom_up":
            self.add_section_bottom_up(**kwargs)

    def add_section_top_down(
        self, **kwargs
    ):
        """
        Sections built from the top down until the bottom of the bottom section
        is equal to the defined string bottom.
        """
        if bool(self.sections) is False:
            temp = 0
            top = self.top
        else:
            temp = len(self.sections)
            top = self.sections[temp - 1]['bottom']

        self.sections[temp] = {}
        self.sections[temp]['top'] = top

        # add the section to the sections dict
        for k, v in kwargs.items():
            self.sections[temp][k] = v

        # sort the dict on depth of tops
        self.sections = {
            k: v
            for k, v in sorted(
                self.sections.items(), key=lambda item: item[1]['top']
            )
        }

        # re-index the keys
        temp = {}
        for i, (k, v) in enumerate(self.sections.items()):
            temp[i] = v

        # check inputs
        for k, v in temp.items():
            if k == 0:
                assert v['top'] == self.top
            else:
                assert v['top'] == temp[k - 1]['bottom']
            assert v['bottom'] <= self.bottom

        if temp[len(temp) - 1]['bottom'] == self.bottom:
            self.complete = True

        self.sections = temp

    def add_section_bottom_up(
        self, **kwargs
    ):
        """
        Sections built from the bottom up until the top of the top
        section is equal to the defined string top.

        Default is to extend the section to the top of the String as
        defined in the String.top property (when length = top = None).

        Parameters:
        -----------

        """
        if bool(self.sections) is False:
            temp = 0
            bottom = self.bottom
        else:
            temp = len(self.sections)
            bottom = self.sections[0]['top']

        if bool(kwargs.get('length')):
            top = bottom - kwargs.get('length')
            if top < self.top:
                top = self.top
                length = bottom - top
        elif bool(kwargs.get('top')):
            length = bottom - kwargs.get('top')
        else:
            top = self.top
            length = self.sections[0]['top']

        self.sections[temp] = {}
        self.sections[temp]['top'] = top
        self.sections[temp]['bottom'] = bottom

        # add the section to the sections dict
        for k, v in kwargs.items():
            self.sections[temp][k] = v

        # sort the dict on depth of tops
        self.sections = {
            k: v
            for k, v in sorted(
                self.sections.items(), key=lambda item: item[1]['bottom']
            )
        }

        # re-index the keys
        temp = {}
        for i, (k, v) in enumerate(self.sections.items()):
            temp[i] = v

        # check inputs
        number_of_sections = len(temp)
        for k, v in temp.items():
            if k == number_of_sections - 1:
                assert v['bottom'] == self.bottom
            else:
                assert v['bottom'] == temp[k + 1]['top']
            assert v['top'] >= self.top

        if temp[0]['top'] == self.top:
            self.complete = True

        self.sections = temp


class WellBore(String):
    def __init__(self, *args, **kwargs):
        """
        Inherits from `String` class, but this makes it more intuitive.
        """
        super().__init__(*args, **kwargs)


class BHA(String):
    def __init__(self, *args, **kwargs):
        """
        Inherits from `String` class, but this makes it more intuitive.
        """
        super().__init__(*args, **kwargs)
